from collections import Counter
from typing import (
    Optional,
    Sequence,
    cast,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.const import (
    PCMK_ROLES_WITH_LEGACY,
    PCMK_STATUS_ROLES,
    PcmkRoleType,
    PcmkStatusRoleType,
)
from pcs.common.pacemaker.role import (
    get_value_primary as get_primary_role_value,
)
from pcs.common.status_dto import (
    AnyResourceStatusDto,
    BundleReplicaStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
    ResourcesStatusDto,
)
from pcs.common.str_tools import format_list
from pcs.lib.pacemaker.values import is_true

_DEFAULT_SEVERITY = reports.ReportItemSeverity.error()
_PRIMITIVE_TAG = "resource"
_GROUP_TAG = "group"
_CLONE_TAG = "clone"
_BUNDLE_TAG = "bundle"
_REPLICA_TAG = "replica"


class ClusterStatusParsingError(Exception):
    def __init__(self, resource_id: str):
        self.resource_id = resource_id


class EmptyResourceIdError(ClusterStatusParsingError):
    def __init__(self) -> None:
        super().__init__("")


class EmptyNodeNameError(ClusterStatusParsingError):
    pass


class UnknownPcmkRoleError(ClusterStatusParsingError):
    def __init__(self, resource_id: str, role: str):
        super().__init__(resource_id)
        self.role = role


class UnexpectedMemberError(ClusterStatusParsingError):
    def __init__(
        self,
        resource_id: str,
        resource_type: str,
        member_id: str,
        member_type: str,
        expected_types: list[str],
    ):
        super().__init__(resource_id)
        self.resource_type = resource_type
        self.member_id = member_id
        self.member_type = member_type
        self.expected_types = expected_types


class MixedMembersError(ClusterStatusParsingError):
    pass


class DifferentMemberIdsError(ClusterStatusParsingError):
    pass


class BundleReplicaMissingImplicitResourceError(ClusterStatusParsingError):
    def __init__(
        self, resource_id: str, replica_id: str, implicit_resource_type: str
    ):
        super().__init__(resource_id)
        self.replica_id = replica_id
        self.implicit_type = implicit_resource_type


class BundleReplicaInvalidMemberCountError(ClusterStatusParsingError):
    def __init__(self, resource_id: str, replica_id: str):
        super().__init__(resource_id)
        self.replica_id = replica_id


class BundleDifferentReplicas(ClusterStatusParsingError):
    pass


class BundleSameIdAsImplicitResourceError(Exception):
    def __init__(self, bundle_id: str, bad_ids: list[str]):
        self.bundle_id = bundle_id
        self.bad_ids = bad_ids


def cluster_status_parsing_error_to_report(
    e: ClusterStatusParsingError,
    severity: reports.ReportItemSeverity = _DEFAULT_SEVERITY,
) -> reports.ReportItem:
    reason = ""
    if isinstance(e, EmptyResourceIdError):
        reason = "Resource with an empty id"
    elif isinstance(e, EmptyNodeNameError):
        reason = (
            f"Resource '{e.resource_id}' contains a node with an empty name"
        )
    elif isinstance(e, UnknownPcmkRoleError):
        reason = (
            f"Resource '{e.resource_id}' contains an unknown role '{e.role}'"
        )
    elif isinstance(e, UnexpectedMemberError):
        reason = (
            f"Unexpected resource '{e.member_id}' of type '{e.member_type}' "
            f"inside of resource '{e.resource_id}' of type '{e.resource_type}'."
            f" Only resources of type {format_list(e.expected_types)} "
            f"can be in a {e.resource_type}"
        )

    elif isinstance(e, MixedMembersError):
        reason = f"Primitive and group members mixed in clone '{e.resource_id}'"
    elif isinstance(e, DifferentMemberIdsError):
        reason = f"Members with different ids in clone '{e.resource_id}'"
    elif isinstance(e, BundleReplicaMissingImplicitResourceError):
        reason = (
            f"Replica '{e.replica_id}' of bundle '{e.resource_id}' "
            f"is missing implicit {e.implicit_type} resource"
        )
    elif isinstance(e, BundleReplicaInvalidMemberCountError):
        reason = (
            f"Replica '{e.replica_id}' of bundle '{e.resource_id}' has "
            "invalid number of members"
        )
    elif isinstance(e, BundleDifferentReplicas):
        reason = f"Replicas of bundle '{e.resource_id}' are not the same"

    return reports.ReportItem(
        severity,
        reports.messages.BadClusterStateData(reason),
    )


def _primitive_to_dto(
    primitive_el: _Element, remove_instance_suffix: bool = False
) -> PrimitiveStatusDto:
    resource_id = _get_resource_id(primitive_el)
    clone_suffix = None
    if remove_instance_suffix:
        resource_id, clone_suffix = _remove_clone_suffix(resource_id)

    role = _get_role(primitive_el)
    target_role = _get_target_role(primitive_el)

    node_names = [
        str(node.attrib["name"]) for node in primitive_el.iterfind("node")
    ]

    if node_names and any(not name for name in node_names):
        raise EmptyNodeNameError(resource_id)

    return PrimitiveStatusDto(
        resource_id,
        clone_suffix,
        str(primitive_el.attrib["resource_agent"]),
        role,
        target_role,
        is_true(primitive_el.get("active", "false")),
        is_true(primitive_el.get("orphaned", "false")),
        is_true(primitive_el.get("blocked", "false")),
        is_true(primitive_el.get("maintenance", "false")),
        primitive_el.get("description"),
        is_true(primitive_el.get("failed", "false")),
        is_true(primitive_el.get("managed", "false")),
        is_true(primitive_el.get("failure_ignored", "false")),
        node_names,
        primitive_el.get("pending"),
        primitive_el.get("locked_to"),
    )


def _group_to_dto(
    group_el: _Element, remove_instance_suffix: bool = False
) -> GroupStatusDto:
    # clone instance id present even when the clone is non unique
    group_id, clone_instance_id = _remove_clone_suffix(
        _get_resource_id(group_el)
    )

    member_list = []

    for member in group_el:
        if member.tag == _PRIMITIVE_TAG:
            member_list.append(
                _primitive_to_dto(member, remove_instance_suffix)
            )
        else:
            raise UnexpectedMemberError(
                group_id,
                "group",
                str(member.attrib["id"]),
                member.tag,
                ["primitive"],
            )

    return GroupStatusDto(
        group_id,
        clone_instance_id,
        is_true(group_el.get("maintenance", "false")),
        group_el.get("description"),
        is_true(group_el.get("managed", "false")),
        is_true(group_el.get("disabled", "false")),
        member_list,
    )


def _clone_to_dto(
    clone_el: _Element, _remove_instance_suffix: bool = False
) -> CloneStatusDto:
    clone_id = _get_resource_id(clone_el)
    is_unique = is_true(clone_el.get("unique", "false"))

    target_role = _get_target_role(clone_el)

    primitive_list = []
    group_list = []

    for member in clone_el:
        if member.tag == _PRIMITIVE_TAG:
            primitive_list.append(_primitive_to_dto(member, is_unique))
        elif member.tag == _GROUP_TAG:
            group_list.append(_group_to_dto(member, is_unique))
        else:
            raise UnexpectedMemberError(
                clone_id,
                "clone",
                str(member.attrib["id"]),
                member.tag,
                ["primitive", "group"],
            )

    if primitive_list and group_list:
        raise MixedMembersError(clone_id)

    if primitive_list:
        if len({res.resource_id for res in primitive_list}) > 1:
            raise DifferentMemberIdsError(clone_id)
    if group_list:
        group_ids = {group.resource_id for group in group_list}
        children_ids = {
            tuple(child.resource_id for child in group.members)
            for group in group_list
        }
        if len(group_ids) > 1 or len(children_ids) > 1:
            raise DifferentMemberIdsError(clone_id)

    return CloneStatusDto(
        clone_id,
        is_true(clone_el.get("multi_state", "false")),
        is_unique,
        is_true(clone_el.get("maintenance", "false")),
        clone_el.get("description"),
        is_true(clone_el.get("managed", "false")),
        is_true(clone_el.get("disabled", "false")),
        is_true(clone_el.get("failed", "false")),
        is_true(clone_el.get("failure_ignored", "false")),
        target_role,
        primitive_list or group_list,
    )


def _bundle_to_dto(
    bundle_el: _Element, _remove_instance_suffix: bool = False
) -> BundleStatusDto:
    bundle_id = _get_resource_id(bundle_el)
    bundle_type = str(bundle_el.attrib["type"])
    is_unique = is_true(bundle_el.get("unique", "false"))

    replica_list = [
        _replica_to_dto(replica, bundle_id, bundle_type, is_unique)
        for replica in bundle_el.iterfind(_REPLICA_TAG)
    ]

    if not _replicas_valid(replica_list):
        raise BundleDifferentReplicas(bundle_id)

    return BundleStatusDto(
        bundle_id,
        bundle_type,
        str(bundle_el.attrib["image"]),
        is_unique,
        is_true(bundle_el.get("maintenance", "false")),
        bundle_el.get("description"),
        is_true(bundle_el.get("managed", "false")),
        is_true(bundle_el.get("failed", "false")),
        replica_list,
    )


class ClusterStatusParser:
    TAG_TO_FUNCTION = {
        _PRIMITIVE_TAG: _primitive_to_dto,
        _GROUP_TAG: _group_to_dto,
        _CLONE_TAG: _clone_to_dto,
        _BUNDLE_TAG: _bundle_to_dto,
    }

    def __init__(self, status: _Element):
        """
        status -- xml element from crm_mon xml, validated using the appropriate
            rng schema
        """
        self._status = status
        self._warnings: reports.ReportItemList = []

    def status_xml_to_dto(self) -> ResourcesStatusDto:
        """
        Return dto containing status of configured resources in the cluster
        """
        resource_list = cast(list[_Element], self._status.xpath("resources/*"))

        resource_dto_list = []
        for resource in resource_list:
            try:
                resource_dto = cast(
                    AnyResourceStatusDto,
                    self.TAG_TO_FUNCTION[resource.tag](resource),
                )
                resource_dto_list.append(resource_dto)
            except BundleSameIdAsImplicitResourceError as e:
                # This is the only error that the user can cause directly by
                # setting the name of the bundle member to be same as one of
                # the implicitly created resource.
                # We only skip such bundles while still providing status of the
                # other resources.
                self._warnings.append(
                    reports.ReportItem.warning(
                        reports.messages.ClusterStatusBundleMemberIdAsImplicit(
                            e.bundle_id, e.bad_ids
                        )
                    )
                )
            except BundleReplicaMissingImplicitResourceError as e:
                # TODO crm_mon on Fedora 39 returns resource_agent in legacy
                # format "ocf::*:*" instead of the new "ocf:*:*" and the parser
                # then cannot find the proper resources in the replicas.
                # Skip bundles when the legacy format is used.
                self._warnings.append(
                    cluster_status_parsing_error_to_report(
                        e, reports.ReportItemSeverity.warning()
                    )
                )

        return ResourcesStatusDto(resource_dto_list)

    def get_warnings(self) -> reports.ReportItemList:
        return self._warnings


def _get_resource_id(resource: _Element) -> str:
    resource_id = resource.attrib["id"]
    if not resource_id:
        raise EmptyResourceIdError()
    return str(resource_id)


def _get_role(resource: _Element) -> PcmkStatusRoleType:
    role = resource.get("role")
    if role is None or role not in PCMK_STATUS_ROLES:
        raise UnknownPcmkRoleError(str(resource.get("id")), str(role))
    return PcmkStatusRoleType(role)


def _get_target_role(resource: _Element) -> Optional[PcmkRoleType]:
    target_role = resource.get("target_role")
    if target_role is None:
        return None
    target_role_normalized = target_role.capitalize()
    if target_role_normalized not in PCMK_ROLES_WITH_LEGACY:
        raise UnknownPcmkRoleError(str(resource.get("id")), target_role)
    return get_primary_role_value(PcmkRoleType(target_role_normalized))


def _remove_clone_suffix(resource_id: str) -> tuple[str, Optional[str]]:
    if ":" in resource_id:
        resource_id, clone_suffix = resource_id.rsplit(":", 1)
        return resource_id, clone_suffix
    return resource_id, None


def _replica_to_dto(
    replica_el: _Element,
    bundle_id: str,
    bundle_type: str,
    remove_instance_suffix: bool = False,
) -> BundleReplicaStatusDto:
    replica_id = str(replica_el.attrib["id"])

    resource_list = [
        _primitive_to_dto(resource, remove_instance_suffix)
        for resource in replica_el.iterfind(_PRIMITIVE_TAG)
    ]

    duplicate_ids = [
        id_
        for id_, count in Counter(
            resource.resource_id for resource in resource_list
        ).items()
        if count > 1
    ]

    if duplicate_ids:
        raise BundleSameIdAsImplicitResourceError(bundle_id, duplicate_ids)

    # TODO pacemaker will probably add prefix
    # "pcmk-internal" to all implicit resources

    container_resource = _pop_implicit_resource(
        resource_list,
        f"{bundle_id}-{bundle_type}-{replica_id}",
        True,
        f"ocf:heartbeat:{bundle_type}",
    )

    if container_resource is None:
        raise BundleReplicaMissingImplicitResourceError(
            bundle_id, replica_id, "container"
        )

    remote_resource = _pop_implicit_resource(
        resource_list, f"{bundle_id}-{replica_id}", True, "ocf:pacemaker:remote"
    )

    # implicit ip address resource might be present
    ip_resource = None
    if (remote_resource is not None and len(resource_list) == 2) or (
        remote_resource is None and len(resource_list) == 1
    ):
        ip_resource = _pop_implicit_resource(
            resource_list, f"{bundle_id}-ip-", False, "ocf:heartbeat:IPaddr2"
        )

    if remote_resource is None and resource_list:
        raise BundleReplicaMissingImplicitResourceError(
            bundle_id, replica_id, "remote"
        )

    member = None
    if remote_resource:
        if len(resource_list) == 1:
            member = resource_list[0]
        else:
            raise BundleReplicaInvalidMemberCountError(bundle_id, replica_id)

    return BundleReplicaStatusDto(
        replica_id,
        member,
        remote_resource,
        container_resource,
        ip_resource,
    )


def _pop_implicit_resource(
    primitive_list: list[PrimitiveStatusDto],
    expected_id: str,
    exact_match: bool,
    resource_agent: str,
) -> Optional[PrimitiveStatusDto]:
    for primitive in primitive_list:
        matching_id = (
            exact_match
            and primitive.resource_id == expected_id
            or not exact_match
            and primitive.resource_id.startswith(expected_id)
        )

        if matching_id and primitive.resource_agent == resource_agent:
            primitive_list.remove(primitive)
            return primitive

    return None


def _replicas_valid(replica_list: Sequence[BundleReplicaStatusDto]) -> bool:
    if not replica_list:
        return True

    member = replica_list[0].member
    ip = replica_list[0].ip_address
    container = replica_list[0].container

    for replica in replica_list:
        if (
            not _cmp_replica_members(member, replica.member, True)
            or not _cmp_replica_members(ip, replica.ip_address, False)
            or not _cmp_replica_members(container, replica.container, False)
        ):
            return False
    return True


def _cmp_replica_members(
    left: Optional[PrimitiveStatusDto],
    right: Optional[PrimitiveStatusDto],
    compare_ids: bool,
) -> bool:
    if left is None and right is None:
        return True
    if left is None:
        return False
    if right is None:
        return False

    if left.resource_agent != right.resource_agent:
        return False

    return not compare_ids or left.resource_id == right.resource_id



from unittest import TestCase

from pcs import settings
from pcs.common import reports
from pcs.common.file_type_codes import (
    COROSYNC_CONF,
    PCS_KNOWN_HOSTS,
    PCS_SETTINGS_CONF,
)
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.lib.commands import pcs_cfgsync as lib

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


class GetConfigs(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def fixture_files_operations(
        self, known_hosts_exists: bool = True, pcs_settings_exists: bool = True
    ) -> None:
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            path=settings.pcsd_known_hosts_location,
            exists=known_hosts_exists,
            name="known-hosts.exists",
        )
        if known_hosts_exists:
            self.config.raw_file.read(
                PCS_KNOWN_HOSTS,
                path=settings.pcsd_known_hosts_location,
                content="known-hosts content".encode("utf-8"),
                name="known-hosts.read",
            )
        self.config.raw_file.exists(
            PCS_SETTINGS_CONF,
            path=settings.pcsd_settings_conf_location,
            exists=pcs_settings_exists,
            name="pcs_settings.exists",
        )
        if pcs_settings_exists:
            self.config.raw_file.read(
                PCS_SETTINGS_CONF,
                path=settings.pcsd_settings_conf_location,
                content="pcs_settings.conf content".encode("utf-8"),
                name="pcs_settings.read",
            )

    def test_bad_cluster_name(self):
        self.config.corosync_conf.load()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_configs(
                self.env_assist.get_env(), "definitely not the right name"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME,
                    cluster_name="definitely not the right name",
                ),
            ]
        )

    def test_not_in_cluster(self):
        # cant read corosync conf
        self.config.corosync_conf.load_content("", exception_msg="some error")
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_configs(self.env_assist.get_env(), "test99"),
            reports=[
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=COROSYNC_CONF,
                    operation="read",
                    reason="some error",
                    file_path=settings.corosync_conf_file,
                )
            ],
            expected_in_processor=False,
        )

    def test_no_files_exist(self):
        self.config.corosync_conf.load()
        self.fixture_files_operations(
            known_hosts_exists=False, pcs_settings_exists=False
        )
        result = lib.get_configs(self.env_assist.get_env(), "test99")
        self.assertEqual(
            SyncConfigsDto(cluster_name="test99", configs={}), result
        )

    def test_some_files_exist(self):
        self.config.corosync_conf.load()
        self.fixture_files_operations(
            known_hosts_exists=True, pcs_settings_exists=False
        )
        result = lib.get_configs(self.env_assist.get_env(), "test99")
        self.assertEqual(
            SyncConfigsDto(
                cluster_name="test99",
                configs={PCS_KNOWN_HOSTS: "known-hosts content"},
            ),
            result,
        )

    def test_all_files_exist(self):
        self.config.corosync_conf.load()
        self.fixture_files_operations()
        result = lib.get_configs(self.env_assist.get_env(), "test99")
        self.assertEqual(
            SyncConfigsDto(
                cluster_name="test99",
                configs={
                    PCS_KNOWN_HOSTS: "known-hosts content",
                    PCS_SETTINGS_CONF: "pcs_settings.conf content",
                },
            ),
            result,
        )

    def test_raw_file_error(self):
        self.config.corosync_conf.load()
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            path=settings.pcsd_known_hosts_location,
            name="known-hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            path=settings.pcsd_known_hosts_location,
            exception_msg="some error",
            name="known-hosts.read",
        )
        self.config.raw_file.exists(
            PCS_SETTINGS_CONF,
            path=settings.pcsd_settings_conf_location,
            name="pcs_settings.exists",
        )
        self.config.raw_file.read(
            PCS_SETTINGS_CONF,
            path=settings.pcsd_settings_conf_location,
            content="pcs_settings.conf content".encode("utf-8"),
            name="pcs_settings.read",
        )

        result = lib.get_configs(self.env_assist.get_env(), "test99")

        self.assertEqual(
            SyncConfigsDto(
                cluster_name="test99",
                configs={PCS_SETTINGS_CONF: "pcs_settings.conf content"},
            ),
            result,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="read",
                    reason="some error",
                    file_path=settings.pcsd_known_hosts_location,
                )
            ]
        )



from pcs import usage
from pcs.cli.cluster_property import command as cluster_property
from pcs.cli.common.errors import raise_command_replaced
from pcs.cli.common.routing import create_router

property_cmd = create_router(
    {
        "help": lambda _lib, _argv, _modifiers: print(
            usage.property_usage(_argv)
        ),
        "set": cluster_property.set_property,
        "unset": cluster_property.unset_property,
        "list": lambda lib, argv, modifiers: raise_command_replaced(
            ["pcs property config"], pcs_version="0.12"
        ),
        "show": lambda lib, argv, modifiers: raise_command_replaced(
            ["pcs property config"], pcs_version="0.12"
        ),
        "config": cluster_property.config,
        "defaults": cluster_property.defaults,
        "describe": cluster_property.describe,
        "get_cluster_properties_definition": (
            cluster_property.print_cluster_properties_definition_legacy
        ),
    },
    ["property"],
    default_cmd="config",
)



import json
from shlex import split
from textwrap import dedent
from unittest import TestCase

from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.resource.clone import CibResourceCloneDto
from pcs.common.pacemaker.resource.group import CibResourceGroupDto
from pcs.common.pacemaker.resource.list import CibResourcesDto

from pcs_test.tools import resources_dto
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class ResourceConfigJson(TestCase):
    def setUp(self):
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-all.xml"),
        )
        self.maxDiff = None

    def test_all(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["resource", "config", "--output-format=json"]
        )
        expected = CibResourcesDto(
            primitives=[
                resources_dto.PRIMITIVE_R1,
                resources_dto.PRIMITIVE_R7,
                resources_dto.PRIMITIVE_R5,
                resources_dto.PRIMITIVE_R2,
                resources_dto.PRIMITIVE_R3,
                resources_dto.PRIMITIVE_R4,
                resources_dto.PRIMITIVE_R6,
            ],
            clones=[
                resources_dto.CLONE_G1,
                resources_dto.CLONE_R6,
            ],
            groups=[
                resources_dto.GROUP_G2,
                resources_dto.GROUP_G1,
            ],
            bundles=[
                resources_dto.BUNDLE_B1,
                resources_dto.BUNDLE_B2,
            ],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_get_multiple(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["resource", "config", "--output-format=json", "G1-clone", "R1"]
        )
        expected = CibResourcesDto(
            primitives=[
                resources_dto.PRIMITIVE_R1,
                resources_dto.PRIMITIVE_R2,
                resources_dto.PRIMITIVE_R3,
                resources_dto.PRIMITIVE_R4,
            ],
            clones=[resources_dto.CLONE_G1],
            groups=[resources_dto.GROUP_G1],
            bundles=[],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)


class ResourceConfigCmdMixin:
    sub_command = None
    expected_json_config_stderr = ""

    def setUp(self):
        self.new_cib_file = get_tmp_file(self._get_tmp_file_name())
        self.pcs_runner_orig = PcsRunner(
            cib_file=get_test_resource("cib-all.xml")
        )
        self.pcs_runner_orig.mock_settings = get_mock_settings()
        self.pcs_runner_new = PcsRunner(cib_file=self.new_cib_file.name)
        self.pcs_runner_new.mock_settings = get_mock_settings()
        write_file_to_tmpfile(
            get_test_resource("cib-empty.xml"), self.new_cib_file
        )
        self.maxDiff = None

    def tearDown(self):
        self.new_cib_file.close()

    def _run_commands(self, stdout):
        cmds = [
            split(cmd)[1:]
            for cmd in stdout.replace("\\\n", "").strip().split(";\n")
        ]
        for cmd in cmds:
            stdout, stderr, retval = self.pcs_runner_new.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )

    def _get_as_json(self, runner):
        stdout, stderr, retval = runner.run(
            [self.sub_command, "config", "--output-format=json"]
        )
        self.assertEqual(stderr, self.expected_json_config_stderr)
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def test_all(self):
        stdout, stderr, retval = self.pcs_runner_orig.run(
            [self.sub_command, "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        self.assertEqual(stderr, "")
        self._run_commands(stdout)
        self.assertEqual(
            self._get_as_json(self.pcs_runner_new),
            self._get_as_json(self.pcs_runner_orig),
        )


class ResourceConfigCmd(ResourceConfigCmdMixin, TestCase):
    sub_command = "resource"

    @staticmethod
    def _get_tmp_file_name():
        return "tier1_resource_test_config_cib.xml"

    def test_unsupported_stonith(self):
        self.pcs_runner_orig = PcsRunner(
            cib_file=get_test_resource("cib-unsupported-stonith-config.xml")
        )
        stdout, stderr, retval = self.pcs_runner_orig.run(
            ["resource", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        self.assertEqual(
            stderr,
            (
                "Warning: Bundle resource 'B1' contains stonith resource: 'S1',"
                " which is unsupported. The bundle resource will be omitted.\n"
                "Warning: Group 'G1' contains stonith resource: 'S2', which is "
                "unsupported. The group will be omitted.\n"
                "Warning: Group 'G2' contains stonith resources: 'S4', 'S5', "
                "which is unsupported. The group will be omitted.\n"
                "Warning: Group 'G3' contains stonith resources: 'S6', 'S7', "
                "which is unsupported. The stonith resources will be omitted.\n"
                "Warning: Group 'G4' contains stonith resource: 'S8', which is "
                "unsupported. The stonith resource will be omitted.\n"
                "Warning: Clone 'S3-clone' contains stonith resource: 'S3', "
                "which is unsupported. The clone will be omitted.\n"
                "Warning: Clone 'G2-clone' contains stonith resources: 'S4', "
                "'S5', which is unsupported. The clone will be omitted.\n"
            ),
        )
        self._run_commands(stdout)
        expected_dict = to_dict(
            CibResourcesDto(
                primitives=[
                    resources_dto.PRIMITIVE_R1,
                    resources_dto.PRIMITIVE_R2,
                    resources_dto.PRIMITIVE_R3,
                ],
                clones=[
                    CibResourceCloneDto(
                        id="G4-clone",
                        description=None,
                        member_id="G4",
                        meta_attributes=[],
                        instance_attributes=[],
                    )
                ],
                groups=[
                    CibResourceGroupDto(
                        id="G3",
                        description=None,
                        member_ids=["R2"],
                        meta_attributes=[],
                        instance_attributes=[],
                    ),
                    CibResourceGroupDto(
                        id="G4",
                        description=None,
                        member_ids=["R3"],
                        meta_attributes=[],
                        instance_attributes=[],
                    ),
                ],
                bundles=[],
            )
        )
        self.assertEqual(
            self._get_as_json(self.pcs_runner_new),
            expected_dict,
        )

    def test_plaintext(self):
        stdout, stderr, retval = self.pcs_runner_orig.run(
            [self.sub_command, "config"]
        )
        self.assertEqual(
            stdout,
            dedent(
                """\
                Resource: R7 (class=ocf provider=pcsmock type=minimal)
                  Description: R7 description is very long " & and special
                  Attributes: R7-instance_attributes
                    envfile=/dev/null
                    fake=looool
                  Meta Attributes: R7-meta_attributes
                    "another one0"="a + b = c"
                    anotherone=something'"special
                    m1=value1
                    m10=value1
                    meta2=valueofmeta2isthisverylongstring
                    meta20=valueofmeta2isthisverylongstring
                  Operations:
                    custom_action: R7-custom_action-interval-10s
                      interval=10s OCF_CHECK_LEVEL=2
                    migrate_from: R7-migrate_from-interval-0s
                      interval=0s timeout=20s
                    migrate_to: R7-migrate_to-interval-0s
                      interval=0s timeout=20s enabled=0 record-pending=0
                    monitor: R7-monitor-interval-10s
                      interval=10s timeout=20s
                    reload: R7-reload-interval-0s
                      interval=0s timeout=20s
                    reload-agent: R7-reload-agent-interval-0s
                      interval=0s timeout=20s
                    start: R7-start-interval-0s
                      interval=0s timeout=20s
                    stop: R7-stop-interval-0s
                      interval=0s timeout=20s
                Group: G2
                  Meta Attributes: G2-meta_attributes
                    meta1=metaval1
                    meta2=metaval2
                  Resource: R5 (class=ocf provider=pcsmock type=minimal)
                    Operations:
                      monitor: R5-monitor-interval-10s
                        interval=10s timeout=20s
                Clone: G1-clone
                  Description: G1-clone description
                  Meta Attributes: G1-clone-meta_attributes
                    promotable=true
                  Group: G1
                    Description: G1 description
                    Resource: R2 (class=ocf provider=pcsmock type=stateful)
                      Operations:
                        monitor: R2-monitor-interval-10s
                          interval=10s timeout=20s
                    Resource: R3 (class=ocf provider=pcsmock type=stateful)
                      Operations:
                        monitor: R3-monitor-interval-10s
                          interval=10s timeout=20s
                    Resource: R4 (class=ocf provider=pcsmock type=stateful)
                      Operations:
                        monitor: R4-monitor-interval-10s
                          interval=10s timeout=20s
                Clone: R6-clone
                  Resource: R6 (class=ocf provider=pcsmock type=minimal)
                    Operations:
                      migrate_from: R6-migrate_from-interval-0s
                        interval=0s timeout=20s
                      migrate_to: R6-migrate_to-interval-0s
                        interval=0s timeout=20s
                      monitor: R6-monitor-interval-10s
                        interval=10s timeout=20s
                      reload: R6-reload-interval-0s
                        interval=0s timeout=20s
                      reload-agent: R6-reload-agent-interval-0s
                        interval=0s timeout=20s
                      start: R6-start-interval-0s
                        interval=0s timeout=20s
                      stop: R6-stop-interval-0s
                        interval=0s timeout=20s
                Bundle: B1
                  Description: B1 description
                  Docker: image=pcs:test replicas=4 replicas-per-host=2 run-command=/bin/true network=extra_network_settings options=extra_options
                  Network: ip-range-start=192.168.100.200 control-port=12345 host-interface=eth0 host-netmask=24
                  Port Mapping:
                    port=1001 (B1-port-map-1001)
                    port=2000 internal-port=2002 (B1-port-map-2000)
                    range=3000-3300 (B1-port-map-3000-3300)
                  Storage Mapping:
                    source-dir=/tmp/docker1a target-dir=/tmp/docker1b (B1-storage-map)
                    source-dir=/tmp/docker2a target-dir=/tmp/docker2b (B1-storage-map-1)
                    source-dir-root=/tmp/docker3a target-dir=/tmp/docker3b (B1-storage-map-2)
                    source-dir-root=/tmp/docker4a target-dir=/tmp/docker4b (B1-storage-map-3)
                  Meta Attributes: B1-meta_attributes
                    is-managed=false
                    target-role=Stopped
                Bundle: B2
                  Docker: image=pcs:test
                  Network: control-port=9000
                  Resource: R1 (class=ocf provider=pcsmock type=minimal)
                    Description: R1 description
                    Operations:
                      monitor: R1-monitor-interval-10s
                        interval=10s timeout=20s
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)



from shlex import quote
from typing import Iterable

from pcs.cli.common.output import (
    INDENT_STEP,
    bool_to_cli_value,
    pairs_to_cmd,
)
from pcs.cli.reports.output import warn
from pcs.common.pacemaker.constraint import (
    CibConstraintOrderAttributesDto,
    CibConstraintOrderDto,
    CibConstraintOrderSetDto,
)
from pcs.common.str_tools import (
    format_optional,
    indent,
    pairs_to_text,
)

from . import set as _set


def _attributes_to_pairs(
    attributes_dto: CibConstraintOrderAttributesDto,
) -> list[tuple[str, str]]:
    pairs = []
    if attributes_dto.symmetrical is not None:
        pairs.append(
            ("symmetrical", bool_to_cli_value(attributes_dto.symmetrical))
        )
    if attributes_dto.require_all is not None:
        pairs.append(
            ("require-all", bool_to_cli_value(attributes_dto.require_all))
        )
    if attributes_dto.score:
        pairs.append(("score", attributes_dto.score))
    if attributes_dto.kind:
        pairs.append(("kind", attributes_dto.kind))
    return pairs


def _attributes_to_text(
    attributes_dto: CibConstraintOrderAttributesDto,
) -> list[str]:
    return pairs_to_text(_attributes_to_pairs(attributes_dto))


def plain_constraint_to_text(
    constraint_dto: CibConstraintOrderDto,
    with_id: bool,
) -> list[str]:
    result = [
        "{first_action}resource '{first_resource}' then {then_action}resource '{then_resource}'".format(
            first_action=format_optional(constraint_dto.first_action),
            first_resource=constraint_dto.first_resource_id,
            then_action=format_optional(constraint_dto.then_action),
            then_resource=constraint_dto.then_resource_id,
        )
    ]
    if with_id:
        result[0] += f" (id: {constraint_dto.attributes.constraint_id})"
    result.extend(
        indent(
            _attributes_to_text(constraint_dto.attributes),
            indent_step=INDENT_STEP,
        )
    )
    return result


def set_constraint_to_text(
    constraint_dto: CibConstraintOrderSetDto,
    with_id: bool,
) -> list[str]:
    return _set.set_constraint_to_text(
        constraint_dto.attributes.constraint_id,
        _attributes_to_text(constraint_dto.attributes),
        constraint_dto.resource_sets,
        with_id,
    )


def constraints_to_text(
    plain_dtos: Iterable[CibConstraintOrderDto],
    set_dtos: Iterable[CibConstraintOrderSetDto],
    with_id: bool,
) -> list[str]:
    result = []
    if plain_dtos:
        result.append("Order Constraints:")
        for constraint_dto in plain_dtos:
            result.extend(
                indent(
                    plain_constraint_to_text(constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    if set_dtos:
        result.append("Order Set Constraints:")
        for set_constraint_dto in set_dtos:
            result.extend(
                indent(
                    set_constraint_to_text(set_constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    return result


def _attributes_to_cmd_pairs(
    attributes_dto: CibConstraintOrderAttributesDto,
) -> list[tuple[str, str]]:
    return [("id", attributes_dto.constraint_id)] + _attributes_to_pairs(
        attributes_dto
    )


def plain_constraint_to_cmd(
    constraint_dto: CibConstraintOrderDto,
) -> list[str]:
    if (
        constraint_dto.first_resource_instance is not None
        or constraint_dto.then_resource_instance is not None
    ):
        warn(
            "Resource instance(s) detected in constraint "
            f"'{constraint_dto.attributes.constraint_id}' but not supported by "
            "this command."
            " Command for creating the constraint is omitted."
        )
        return []
    result = [
        "pcs -- constraint order {first_action}{first_resource_id} then {then_action}{then_resource_id}".format(
            first_action=format_optional(constraint_dto.first_action),
            first_resource_id=quote(constraint_dto.first_resource_id),
            then_action=format_optional(constraint_dto.then_action),
            then_resource_id=quote(constraint_dto.then_resource_id),
        )
    ]
    params = pairs_to_cmd(_attributes_to_cmd_pairs(constraint_dto.attributes))
    if params:
        result.extend(indent([params], indent_step=INDENT_STEP))
    return result


def set_constraint_to_cmd(
    constraint_dto: CibConstraintOrderSetDto,
) -> list[str]:
    result = ["pcs -- constraint order"]
    for resource_set in constraint_dto.resource_sets:
        set_cmd_part = _set.resource_set_to_cmd(resource_set)
        if not set_cmd_part:
            return []
        result.extend(indent(set_cmd_part, indent_step=INDENT_STEP))
    pairs = []
    for pair in _attributes_to_cmd_pairs(constraint_dto.attributes):
        # this list is based on pcs.lib.cib.constraint.order.ATTRIB
        if pair[0] not in ("symmetrical", "kind", "id"):
            warn(
                f"Option '{pair[0]}' detected in constraint "
                f"'{constraint_dto.attributes.constraint_id}' but not "
                "supported by this command."
                " Command for creating the constraint is omitted."
            )
            return []
        pairs.append(pair)
    if pairs:
        result.extend(
            indent(
                [f"setoptions {pairs_to_cmd(pairs)}"], indent_step=INDENT_STEP
            )
        )
    return result



from dataclasses import dataclass
from typing import Mapping

from pcs.common.host import PcsKnownHost


@dataclass(frozen=True)
class KnownHosts:
    format_version: int
    data_version: int
    known_hosts: Mapping[str, PcsKnownHost]



from unittest import TestCase

from lxml import etree

from pcs.common import (
    const,
    reports,
)
from pcs.common.pacemaker.types import CibResourceDiscovery
from pcs.common.types import CibRuleExpressionType
from pcs.lib.cib import const as cib_const
from pcs.lib.cib.constraint import location
from pcs.lib.cib.rule import parse_rule
from pcs.lib.cib.tools import IdProvider
from pcs.lib.xml_tools import etree_to_str

from pcs_test.tier0.lib.cib.constraint.test_common import (
    DuplicatesCheckerTestBase,
)
from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)


class IsLocationConstraint(TestCase):
    def test_location_constraint_true(self):
        self.assertTrue(
            location.is_location_constraint(etree.Element("rsc_location"))
        )

    def test_location_constraint_false(self):
        self.assertFalse(
            location.is_location_constraint(etree.Element("location"))
        )


class IsLocationConstraintWithRule(TestCase):
    def test_location_with_rule(self):
        self.assertTrue(
            location.is_location_constraint_with_rule(
                etree.fromstring("<rsc_location><rule/></rsc_location>")
            )
        )

    def test_location_no_rule(self):
        self.assertFalse(
            location.is_location_constraint_with_rule(
                etree.fromstring("<rsc_location />")
            )
        )

    def test_element_with_rule(self):
        self.assertFalse(
            location.is_location_constraint_with_rule(
                etree.fromstring("<element><rule/></element>")
            )
        )


class IsLocationRule(TestCase):
    def test_location_constraint_true(self):
        self.assertTrue(
            location.is_location_rule(
                etree.SubElement(etree.Element("rsc_location"), "rule")
            )
        )

    def test_location_constraint_false(self):
        for element in (
            etree.Element("rule"),
            etree.SubElement(etree.Element("rsc_order"), "rule"),
            etree.SubElement(
                etree.SubElement(etree.Element("rsc_location"), "rsc_order"),
                "rule",
            ),
        ):
            with self.subTest(element=element):
                self.assertFalse(location.is_location_rule(element))


class DuplicatesCheckerLocationRulePlainTest(DuplicatesCheckerTestBase):
    cib = etree.fromstring(
        """
        <constraints>
          <rsc_location id="L1" rsc="R1" node="node1" score="100" />
          <rsc_location id="L2" rsc="R1">
            <rule id="L2-rule" boolean-op="and" score="0">
              <expression id="L2-rule-expr-1"
                  attribute="foo" operation="gt" type="version" value="1.2"
              />
              <expression id="L2-rule-expr"
                  attribute="#uname" operation="eq" value="node3"
              />
            </rule>
            <rule id="L2-rule-1" boolean-op="and" score="0">
              <date_expression id="L2-rule-1-expr" operation="date_spec">
                <date_spec id="L2-rule-1-expr-datespec" weekdays="1-5" />
              </date_expression>
            </rule>
          </rsc_location>
          <rsc_location id="L3" rsc="R1">
            <rule id="L3-rule" boolean-op="and" score="0">
              <expression id="L3-rule-expr"
                  attribute="#uname" operation="eq" type="string" value="node3"
              />
              <expression id="L3-rule-expr-1"
                  attribute="foo" operation="gt" type="version" value="1.2"
              />
            </rule>
          </rsc_location>
          <rsc_location id="L4" rsc="R1">
            <rule id="L4-rule" boolean-op="and" score="0">
              <date_expression id="L4-rule-expr" operation="date_spec">
                <date_spec id="L4-rule-expr-datespec" weekdays="1-5" />
              </date_expression>
            </rule>
          </rsc_location>
          <rsc_location id="L5" rsc="R2">
            <rule id="L5-rule" boolean-op="and" score="0">
              <date_expression id="L5-rule-expr" operation="date_spec">
                <date_spec id="L5-rule-expr-datespec" weekdays="1-5" />
              </date_expression>
            </rule>
          </rsc_location>
          <rsc_location id="L6" rsc-pattern="R*">
            <rule id="L6-rule" boolean-op="and" score="0">
              <expression id="L6-rule-expr"
                  attribute="#uname" operation="eq" value="node3"
              />
            </rule>
          </rsc_location>
          <rsc_location id="L7" rsc-pattern="R*">
            <rule id="L7-rule" boolean-op="and" score="0">
              <expression id="L7-rule-expr"
                  attribute="#uname" operation="eq" type="string" value="node3"
              />
            </rule>
          </rsc_location>
        </constraints>
        """
    )

    def test_success(self):
        duplicates = {
            # L1 has no rules
            # L2 has more than one rule - not compatible with the checker
            "L3": ["L2"],  # rsc matches, the only rule matches when normalized
            "L4": ["L2"],  # rsc matches, the only rule matches
            "L5": [],  # rsc doesn't match, rule matches
            "L6": ["L7"],  # rsc-pattern matches, normalized rule matches
            "L7": ["L6"],  # rsc-pattern matches, normalized rule matches
        }
        checker = location.DuplicatesCheckerLocationRulePlain()
        self.assert_success(self.cib, checker, duplicates)


class ValidateCreatePlainWithRuleCommonMixin:
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <cib>
              <configuration>
                <resources>
                  <primitive id="R1" class="ocf" provider="pacemaker" type="Dummy">
                    <operations>
                      <op id="R1-monitor" name="monitor" interval="60s" />
                    </operations>
                  </primitive>
                  <clone id="C1">
                    <primitive id="C1R1" class="ocf" provider="pacemaker" type="Dummy" />
                  </clone>
                </resources>
                <constraints />
              </configuration>
            </cib>
            """
        )
        self.resource = self.cib.xpath(".//primitive[@id='R1']")[0]
        self.id_provider = IdProvider(self.cib)

    def get_validator(
        self,
        id_provider,
        rule_str,
        rule_options,
        constraint_options=None,
        constrained_el=None,
    ):
        return location.ValidateCreatePlainWithRule(
            id_provider,
            rule_str,
            rule_options,
            constraint_options or {},
            constrained_el if constrained_el is not None else self.resource,
        )

    def test_success_minimal(self):
        validator = self.get_validator(self.id_provider, "#uname eq node1", {})
        assert_report_item_list_equal(validator.validate(), [])

    def test_success_rule_options_1(self):
        validator = self.get_validator(
            self.id_provider,
            "#uname eq node1",
            {
                "id": "ruleid",
                "role": const.PCMK_ROLE_PROMOTED,
                "score": "123",
            },
        )
        assert_report_item_list_equal(validator.validate(), [])

    def test_success_rule_options_2(self):
        validator = self.get_validator(
            self.id_provider,
            "#uname eq node1",
            {
                "id": "ruleid",
                "role": const.PCMK_ROLE_UNPROMOTED,
                "score-attribute": "something",
            },
        )
        assert_report_item_list_equal(validator.validate(), [])

    def test_refuse_legacy_role(self):
        validator = self.get_validator(
            self.id_provider,
            "#uname eq node1",
            {
                "role": const.PCMK_ROLE_PROMOTED_LEGACY,
            },
        )
        assert_report_item_list_equal(
            validator.validate(),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value=const.PCMK_ROLE_PROMOTED_LEGACY,
                    allowed_values=(
                        const.PCMK_ROLE_PROMOTED,
                        const.PCMK_ROLE_UNPROMOTED,
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_rule_options_bad(self):
        for force_flags in ({}, {reports.codes.FORCE}):
            with self.subTest(force_flags=force_flags):
                validator = self.get_validator(
                    self.id_provider,
                    "#uname eq node1",
                    {
                        "id": "bad id",
                        "role": "bad role",
                        "score": "bad score",
                        "score-attribute": "something",
                        "bad": "option",
                    },
                )
                assert_report_item_list_equal(
                    validator.validate(force_flags),
                    [
                        fixture.error(
                            reports.codes.INVALID_OPTIONS,
                            option_names=["bad"],
                            allowed=["id", "role", "score", "score-attribute"],
                            option_type="rule",
                            allowed_patterns=[],
                        ),
                        fixture.error(
                            reports.codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                            option_names=["score", "score-attribute"],
                            option_type="rule",
                        ),
                        fixture.error(
                            reports.codes.INVALID_ID_BAD_CHAR,
                            id="bad id",
                            id_description="rule id",
                            invalid_character=" ",
                            is_first_char=False,
                        ),
                        fixture.error(
                            reports.codes.INVALID_SCORE,
                            score="bad score",
                        ),
                        fixture.error(
                            reports.codes.INVALID_OPTION_VALUE,
                            option_name="role",
                            option_value="bad role",
                            allowed_values=(
                                const.PCMK_ROLE_PROMOTED,
                                const.PCMK_ROLE_UNPROMOTED,
                            ),
                            cannot_be_empty=False,
                            forbidden_characters=None,
                        ),
                    ],
                )

    def test_rule_not_specified(self):
        for force_flags in ({}, {reports.codes.FORCE}):
            with self.subTest(force_flags=force_flags):
                validator = self.get_validator(self.id_provider, "", {})
                assert_report_item_list_equal(
                    validator.validate(force_flags),
                    [
                        fixture.error(
                            reports.codes.RULE_NO_EXPRESSION_SPECIFIED
                        ),
                    ],
                )

    def test_rule_not_valid(self):
        for force_flags in ({}, {reports.codes.FORCE}):
            with self.subTest(force_flags=force_flags):
                validator = self.get_validator(
                    self.id_provider, "#uname eq", {}
                )
                assert_report_item_list_equal(
                    validator.validate(force_flags),
                    [
                        fixture.error(
                            reports.codes.RULE_EXPRESSION_PARSE_ERROR,
                            rule_string="#uname eq",
                            reason="Expected <attribute value>",
                            rule_line="#uname eq",
                            line_number=1,
                            column_number=10,
                            position=9,
                        ),
                    ],
                )

    def test_rule_with_rsc_expression(self):
        for force_flags in ({}, {reports.codes.FORCE}):
            with self.subTest(force_flags=force_flags):
                validator = self.get_validator(
                    self.id_provider, "#uname eq node1 or resource ocf::", {}
                )
                assert_report_item_list_equal(
                    validator.validate(force_flags),
                    [
                        fixture.error(
                            reports.codes.RULE_EXPRESSION_NOT_ALLOWED,
                            expression_type=CibRuleExpressionType.RSC_EXPRESSION,
                        )
                    ],
                )

    def test_rule_with_op_expression(self):
        for force_flags in ({}, {reports.codes.FORCE}):
            with self.subTest(force_flags=force_flags):
                validator = self.get_validator(
                    self.id_provider, "#uname eq node1 or op moonitor", {}
                )
                assert_report_item_list_equal(
                    validator.validate(force_flags),
                    [
                        fixture.error(
                            reports.codes.RULE_EXPRESSION_NOT_ALLOWED,
                            expression_type=CibRuleExpressionType.OP_EXPRESSION,
                        )
                    ],
                )

    def test_success_constraint_options(self):
        validator = location.ValidateCreatePlainWithRule(
            self.id_provider,
            "#uname eq node1",
            {},
            {"id": "location-1", "resource-discovery": "always"},
            self.resource,
        )
        assert_report_item_list_equal(validator.validate(), [])

    def test_constraint_id_already_exists(self):
        test_id = "location-1"
        self.id_provider.book_ids(test_id)
        for force_flags in ({}, {reports.codes.FORCE}):
            with self.subTest(force_flags=force_flags):
                validator = location.ValidateCreatePlainWithRule(
                    self.id_provider,
                    "#uname eq node1",
                    {},
                    {"id": test_id},
                    self.resource,
                )
                assert_report_item_list_equal(
                    validator.validate(force_flags),
                    [
                        fixture.error(
                            reports.codes.ID_ALREADY_EXISTS,
                            id=test_id,
                        )
                    ],
                )

    def test_resource_discovery_invalid(self):
        validator = location.ValidateCreatePlainWithRule(
            self.id_provider,
            "#uname eq node1",
            {},
            {"resource-discovery": "bad value"},
            self.resource,
        )
        assert_report_item_list_equal(
            validator.validate(),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    force_code=reports.codes.FORCE,
                    option_name="resource-discovery",
                    option_value="bad value",
                    allowed_values=[
                        CibResourceDiscovery.ALWAYS,
                        CibResourceDiscovery.EXCLUSIVE,
                        CibResourceDiscovery.NEVER,
                    ],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_resource_discovery_invalid_forced(self):
        validator = location.ValidateCreatePlainWithRule(
            self.id_provider,
            "#uname eq node1",
            {},
            {"resource-discovery": "bad value"},
            self.resource,
        )
        assert_report_item_list_equal(
            validator.validate({reports.codes.FORCE}),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="resource-discovery",
                    option_value="bad value",
                    allowed_values=[
                        CibResourceDiscovery.ALWAYS,
                        CibResourceDiscovery.EXCLUSIVE,
                        CibResourceDiscovery.NEVER,
                    ],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )


class ValidateCreatePlainWithRuleWithResource(
    ValidateCreatePlainWithRuleCommonMixin, TestCase
):
    def setUp(self):
        super().setUp()
        self.resource_in_clone = self.cib.xpath(".//primitive[@id='C1R1']")[0]

    def test_resource_in_clone(self):
        validator = location.ValidateCreatePlainWithRule(
            self.id_provider,
            "#uname eq node1",
            {},
            {},
            self.resource_in_clone,
        )
        assert_report_item_list_equal(
            validator.validate(),
            [
                fixture.error(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    force_code=reports.codes.FORCE,
                    resource_id="C1R1",
                    parent_type="clone",
                    parent_id="C1",
                )
            ],
        )

    def test_resource_in_clone_forced(self):
        validator = location.ValidateCreatePlainWithRule(
            self.id_provider,
            "#uname eq node1",
            {},
            {},
            self.resource_in_clone,
        )
        assert_report_item_list_equal(
            validator.validate({reports.codes.FORCE}),
            [
                fixture.warn(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    resource_id="C1R1",
                    parent_type="clone",
                    parent_id="C1",
                )
            ],
        )

    def test_not_resource(self):
        element = self.cib.xpath(".//*[@id='R1-monitor']")[0]
        validator = location.ValidateCreatePlainWithRule(
            self.id_provider, "#uname eq node1", {}, {}, element
        )
        assert_report_item_list_equal(
            validator.validate(),
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R1-monitor",
                    expected_types=sorted(cib_const.TAG_LIST_CONSTRAINABLE),
                    current_type="op",
                )
            ],
        )


class ValidateCreatePlainWithRuleWithoutResource(
    ValidateCreatePlainWithRuleCommonMixin, TestCase
):
    def setUp(self):
        super().setUp()
        self.resource = None


class CreatePlainWithRule(TestCase):
    @staticmethod
    def fixture_cib(constraint=""):
        return f"""
            <cib>
              <resources>
                <primitive id="R1" class="ocf" provider="pacemaker" type="Dummy" />
              </resources>
              <constraints>{constraint}</constraints>
            </cib>
        """

    @staticmethod
    def run_create_constraint(
        cib,
        resource_id_type,
        resource_id,
        rule_options,
        constraint_options,
        rule="#uname eq node1",
    ):
        element = cib.xpath(".//constraints")[0]
        id_provider = IdProvider(cib)
        rule_tree = parse_rule(rule)
        location.create_plain_with_rule(
            element,
            id_provider,
            const.PCMK_NEW_ROLES_CIB_VERSION,
            resource_id_type,
            resource_id,
            rule_tree,
            rule_options,
            constraint_options,
        )

    def test_minimal_resource(self):
        cib = etree.fromstring(self.fixture_cib())
        cib_expected = self.fixture_cib(
            """
                <rsc_location id="location-R1" rsc="R1">
                  <rule id="location-R1-rule" boolean-op="and" score="INFINITY">
                    <expression id="location-R1-rule-expr"
                        attribute="#uname" operation="eq" value="node1"
                    />
                  </rule>
                </rsc_location>
            """
        )
        self.run_create_constraint(
            cib, const.RESOURCE_ID_TYPE_PLAIN, "R1", {}, {}
        )
        assert_xml_equal(cib_expected, etree_to_str(cib))

    def test_minimal_resource_pattern(self):
        cib = etree.fromstring(self.fixture_cib())
        cib_expected = self.fixture_cib(
            """
                <rsc_location id="location-Rd" rsc-pattern="R\\d+">
                  <rule id="location-Rd-rule" boolean-op="and" score="INFINITY">
                    <expression id="location-Rd-rule-expr"
                        attribute="#uname" operation="eq" value="node1"
                    />
                  </rule>
                </rsc_location>
            """
        )
        self.run_create_constraint(
            cib, const.RESOURCE_ID_TYPE_REGEXP, "R\\d+", {}, {}
        )
        assert_xml_equal(cib_expected, etree_to_str(cib))

    def test_options(self):
        cib = etree.fromstring(self.fixture_cib())
        cib_expected = self.fixture_cib(
            f"""
                <rsc_location id="L1" rsc="R1" resource-discovery="exclusive">
                  <rule id="X1" boolean-op="and" score="123"
                      role="{const.PCMK_ROLE_UNPROMOTED}"
                  >
                    <expression id="X1-expr"
                        attribute="#uname" operation="eq" value="node1"
                    />
                  </rule>
                </rsc_location>
            """
        )
        self.run_create_constraint(
            cib,
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            {"id": "X1", "score": "123", "role": const.PCMK_ROLE_UNPROMOTED},
            {"id": "L1", "resource-discovery": CibResourceDiscovery.EXCLUSIVE},
        )
        assert_xml_equal(cib_expected, etree_to_str(cib))

    def test_score_attribute(self):
        cib = etree.fromstring(self.fixture_cib())
        cib_expected = self.fixture_cib(
            """
                <rsc_location id="location-R1" rsc="R1">
                  <rule id="location-R1-rule" boolean-op="and"
                    score-attribute="something"
                >
                    <expression id="location-R1-rule-expr"
                        attribute="#uname" operation="eq" value="node1"
                    />
                  </rule>
                </rsc_location>
            """
        )
        self.run_create_constraint(
            cib,
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            {"score-attribute": "something"},
            {},
        )
        assert_xml_equal(cib_expected, etree_to_str(cib))



from dataclasses import replace as dc_replace

from pcs.common.resource_agent.dto import (
    ResourceAgentParameterDto,
    ResourceMetaAttributesMetadataDto,
)
from pcs.lib.resource_agent import const as ra_const

_ADDITIONAL_FENCING_META_ATTRIBUTES_DTO = [
    ResourceAgentParameterDto(
        name="provides",
        shortdesc=None,
        longdesc="Any special capability provided by the fence device.",
        type="string",
        default=None,
        enum_values=None,
        required=False,
        advanced=False,
        deprecated=False,
        deprecated_by=[],
        deprecated_desc=None,
        unique_group=None,
        reloadable=False,
    )
]


def get_fixture_meta_attributes_dto(
    agent_name=ra_const.PRIMITIVE_META,
) -> ResourceMetaAttributesMetadataDto:
    metadata_dto = ResourceMetaAttributesMetadataDto(
        name=agent_name,
        parameters=[
            ResourceAgentParameterDto(
                name="priority",
                shortdesc="Resource assignment priority",
                longdesc=(
                    "Resource assignment priority.\nIf not all resources can "
                    "be active, the cluster will stop lower-priority "
                    "resources in order to keep higher-priority ones active."
                ),
                type="score",
                default="0",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="critical",
                shortdesc=(
                    "Default value for influence in colocation constraints"
                ),
                longdesc=(
                    "Default value for influence in colocation constraints.\n"
                    "Use this value as the default for influence in all "
                    "colocation constraints involving this resource, as well "
                    "as in the implicit colocation constraints created if "
                    "this resource is in a group."
                ),
                type="boolean",
                default="true",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="target-role",
                shortdesc=(
                    "State the cluster should attempt to keep this resource in"
                ),
                longdesc=(
                    "State the cluster should attempt to keep this resource "
                    'in.\n"Stopped" forces the resource to be stopped. '
                    '"Started" allows the resource to be started (and in the '
                    "case of promotable clone resources, promoted if "
                    'appropriate). "Unpromoted" allows the resource to be '
                    "started, but only in the unpromoted role if the resource "
                    'is promotable. "Promoted" is equivalent to "Started".'
                ),
                type="select",
                default="Started",
                enum_values=[
                    "Stopped",
                    "Started",
                    "Unpromoted",
                    "Promoted",
                ],
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="is-managed",
                shortdesc=(
                    "Whether the cluster is allowed to actively change the "
                    "resource's state"
                ),
                longdesc=(
                    "Whether the cluster is allowed to actively change the "
                    "resource's state.\nIf false, the cluster will not start, "
                    "stop, promote, or demote the resource on any node. "
                    "Recurring actions for the resource are unaffected. If "
                    "true, a true value for the maintenance-mode cluster "
                    "option, the maintenance node attribute, or the "
                    "maintenance resource meta-attribute overrides this."
                ),
                type="boolean",
                default="true",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="maintenance",
                shortdesc=(
                    "If true, the cluster will not schedule any actions "
                    "involving the resource"
                ),
                longdesc=(
                    "If true, the cluster will not schedule any actions "
                    "involving the resource.\nIf true, the cluster will not "
                    "start, stop, promote, or demote the resource on any node, "
                    "and will pause any recurring monitors (except those "
                    'specifying role as "Stopped"). If false, a true value '
                    "for the maintenance-mode cluster option or maintenance "
                    "node attribute overrides this."
                ),
                type="boolean",
                default="false",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="resource-stickiness",
                shortdesc=(
                    "Score to add to the current node when a resource is "
                    "already active"
                ),
                longdesc=(
                    "Score to add to the current node when a resource is "
                    "already active. This allows running resources to stay "
                    "where they are, even if they would be placed elsewhere if "
                    "they were being started from a stopped state. The default "
                    "is 1 for individual clone instances, and 0 for all other "
                    "resources."
                ),
                type="score",
                default=None,
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="requires",
                shortdesc="Conditions under which the resource can be started",
                longdesc=(
                    "Conditions under which the resource can "
                    'be started. "nothing" means the cluster '
                    'can always start this resource. "quorum" '
                    "means the cluster can start this resource "
                    "only if a majority of the configured "
                    'nodes are active. "fencing" means the '
                    "cluster can start this resource only if a "
                    "majority of the configured nodes are "
                    "active and any failed or unknown nodes "
                    'have been fenced. "unfencing" means the '
                    "cluster can start this resource only if a "
                    "majority of the configured nodes are "
                    "active and any failed or unknown nodes "
                    "have been fenced, and only on nodes that "
                    "have been unfenced. The default is "
                    '"quorum" for resources with a class of '
                    'stonith; otherwise, "unfencing" if '
                    "unfencing is active in the cluster; "
                    'otherwise, "fencing" if the '
                    "fencing-enabled cluster option is true; "
                    'otherwise, "quorum".'
                ),
                type="select",
                default=None,
                enum_values=[
                    "nothing",
                    "quorum",
                    "fencing",
                    "unfencing",
                ],
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="migration-threshold",
                shortdesc=(
                    "Number of failures on a node before the resource becomes "
                    "ineligible to run there."
                ),
                longdesc=(
                    "Number of failures on a node before the "
                    "resource becomes ineligible to run "
                    "there.\n"
                    "Number of failures that may occur for "
                    "this resource on a node, before that node "
                    "is marked ineligible to host this "
                    "resource. A value of 0 indicates that "
                    "this feature is disabled (the node will "
                    "never be marked ineligible). By contrast, "
                    'the cluster treats "INFINITY" (the '
                    "default) as a very large but finite "
                    "number. This option has an effect only if "
                    "the failed operation specifies its "
                    'on-fail attribute as "restart" (the '
                    "default), and additionally for failed "
                    "start operations, if the "
                    "start-failure-is-fatal cluster property "
                    "is set to false."
                ),
                type="score",
                default="INFINITY",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="failure-timeout",
                shortdesc=(
                    "Number of seconds before acting as if a failure had not "
                    "occurred"
                ),
                longdesc=(
                    "Number of seconds before acting as if a failure had not "
                    "occurred.\nNumber of seconds after a failed action for "
                    "this resource before acting as if the failure had not "
                    "occurred, and potentially allowing the resource back to "
                    "the node on which it failed. A value of 0 indicates that "
                    "this feature is disabled."
                ),
                type="duration",
                default="0",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="multiple-active",
                shortdesc=(
                    "What to do if the cluster finds the resource active on "
                    "more than one node"
                ),
                longdesc=(
                    "What to do if the cluster finds the "
                    "resource active on more than one node. "
                    '"block" means to mark the resource as '
                    'unmanaged. "stop_only" means to stop all '
                    "active instances of this resource and "
                    'leave them stopped. "stop_start" means to '
                    "stop all active instances of this "
                    "resource and start the resource in one "
                    'location only. "stop_unexpected" means to '
                    "stop all active instances of this "
                    "resource except where the resource should "
                    "be active. (This should be used only when "
                    "extra instances are not expected to "
                    "disrupt existing instances, and the "
                    "resource agent's monitor of an existing "
                    "instance is capable of detecting any "
                    "problems that could be caused. Note that "
                    "any resources ordered after this one will "
                    "still need to be restarted.)"
                ),
                type="select",
                default="stop_start",
                enum_values=[
                    "block",
                    "stop_only",
                    "stop_start",
                    "stop_unexpected",
                ],
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="allow-migrate",
                shortdesc=(
                    'Whether the cluster should try to "live '
                    'migrate" this resource when it needs to '
                    "be moved"
                ),
                longdesc=(
                    'Whether the cluster should try to "live '
                    'migrate" this resource when it needs to '
                    "be moved. The default is true for "
                    "ocf:pacemaker:remote resources, and false "
                    "otherwise."
                ),
                type="boolean",
                default=None,
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="allow-unhealthy-nodes",
                shortdesc=(
                    "Whether the resource should be allowed "
                    "to run on a node even if the node's "
                    "health score would otherwise prevent it"
                ),
                longdesc=None,
                type="boolean",
                default="false",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="container-attribute-target",
                shortdesc="Where to check user-defined node attributes",
                longdesc=(
                    "Where to check user-defined node "
                    "attributes.\n"
                    "Whether to check user-defined node "
                    "attributes on the physical host where a "
                    "container is running or on the local "
                    "node. This is usually set for a bundle "
                    "resource and inherited by the bundle's "
                    'primitive resource. A value of "host" '
                    "means to check user-defined node "
                    "attributes on the underlying physical "
                    "host. Any other value means to check "
                    "user-defined node attributes on the local "
                    "node (for a bundled primitive resource, "
                    "this is the bundle node)."
                ),
                type="string",
                default=None,
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="remote-node",
                shortdesc=(
                    "Name of the Pacemaker Remote guest node "
                    "this resource is associated with, if any"
                ),
                longdesc=(
                    "Name of the Pacemaker Remote guest node "
                    "this resource is associated with, if any. "
                    "If specified, this both enables the "
                    "resource as a guest node and defines the "
                    "unique name used to identify the guest "
                    "node. The guest must be configured to run "
                    "the Pacemaker Remote daemon when it is "
                    "started. WARNING: This value cannot "
                    "overlap with any resource or node IDs."
                ),
                type="string",
                default=None,
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="remote-addr",
                shortdesc=(
                    "If remote-node is specified, the IP address or hostname "
                    "used to connect to the guest via Pacemaker Remote"
                ),
                longdesc=(
                    "If remote-node is specified, the IP address or hostname "
                    "used to connect to the guest via Pacemaker Remote. The "
                    "Pacemaker Remote daemon on the guest must be configured "
                    "to accept connections on this address. The default is "
                    "the value of the remote-node meta-attribute."
                ),
                type="string",
                default=None,
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="remote-port",
                shortdesc=(
                    "If remote-node is specified, port on the guest used for "
                    "its Pacemaker Remote connection"
                ),
                longdesc=(
                    "If remote-node is specified, port on the guest used for "
                    "its Pacemaker Remote connection.\nIf remote-node is "
                    "specified, the port on the guest used for its Pacemaker "
                    "Remote connection. The Pacemaker Remote daemon on the "
                    "guest must be configured to listen "
                    "on this port."
                ),
                type="port",
                default="3121",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="remote-connect-timeout",
                shortdesc=(
                    "If remote-node is specified, how long before a pending "
                    "Pacemaker Remote guest connection times out."
                ),
                longdesc=None,
                type="timeout",
                default="60s",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
            ResourceAgentParameterDto(
                name="remote-allow-migrate",
                shortdesc=(
                    "If remote-node is specified, this acts as the "
                    "allow-migrate meta-attribute for the implicit remote "
                    "connection resource (ocf:pacemaker:remote)."
                ),
                longdesc=None,
                type="boolean",
                default="true",
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            ),
        ],
    )
    if agent_name == ra_const.STONITH_META:
        metadata_dto = dc_replace(
            metadata_dto,
            parameters=(
                metadata_dto.parameters
                + _ADDITIONAL_FENCING_META_ATTRIBUTES_DTO
            ),
        )
    return metadata_dto


FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META = sorted(
    parameter.name
    for parameter in get_fixture_meta_attributes_dto(
        ra_const.PRIMITIVE_META
    ).parameters
)

FIXTURE_KNOWN_META_NAMES_STONITH_META = sorted(
    parameter.name
    for parameter in get_fixture_meta_attributes_dto(
        ra_const.STONITH_META
    ).parameters
)



from shlex import quote
from typing import (
    Iterable,
    Optional,
)

from pcs.cli.common.output import (
    INDENT_STEP,
    pairs_to_cmd,
)
from pcs.cli.reports.output import warn
from pcs.cli.rule import rule_expression_dto_to_lines
from pcs.common.pacemaker.constraint import (
    CibConstraintColocationAttributesDto,
    CibConstraintColocationDto,
    CibConstraintColocationSetDto,
)
from pcs.common.str_tools import (
    format_name_value_list,
    format_optional,
    indent,
)
from pcs.common.types import StringCollection

from . import set as _set


def _attributes_to_pairs(
    attributes_dto: CibConstraintColocationAttributesDto,
) -> list[tuple[str, str]]:
    pairs = []
    if attributes_dto.score:
        pairs.append(("score", attributes_dto.score))
    if attributes_dto.influence:
        pairs.append(("influence", attributes_dto.influence))
    return pairs


def _attributes_to_text(
    attributes_dto: CibConstraintColocationAttributesDto,
    with_id: bool,
    extra_attributes: Iterable[tuple[str, str]] = (),
) -> list[str]:
    result = [
        " ".join(
            format_name_value_list(
                _attributes_to_pairs(attributes_dto) + list(extra_attributes)
            )
        )
    ]
    if attributes_dto.lifetime:
        result.append("Lifetime:")
        for rule_dto in attributes_dto.lifetime:
            result.extend(
                indent(
                    rule_expression_dto_to_lines(rule_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )

    return result


def plain_constraint_to_text(
    constraint_dto: CibConstraintColocationDto,
    with_id: bool,
) -> list[str]:
    result = [
        "{resource_role}resource '{resource_id}' with {with_resource_role}resource '{with_resource_id}'".format(
            resource_role=format_optional(constraint_dto.resource_role),
            resource_id=constraint_dto.resource_id,
            with_resource_role=format_optional(
                constraint_dto.with_resource_role
            ),
            with_resource_id=constraint_dto.with_resource_id,
        )
    ]
    if with_id:
        result[0] += f" (id: {constraint_dto.attributes.constraint_id})"
    extra_attributes = []
    if constraint_dto.node_attribute:
        extra_attributes += [
            ("node-attribute", constraint_dto.node_attribute),
        ]
    result.extend(
        indent(
            _attributes_to_text(
                constraint_dto.attributes, with_id, extra_attributes
            ),
            indent_step=INDENT_STEP,
        )
    )
    return result


def set_constraint_to_text(
    constraint_dto: CibConstraintColocationSetDto,
    with_id: bool,
) -> list[str]:
    return _set.set_constraint_to_text(
        constraint_dto.attributes.constraint_id,
        _attributes_to_text(constraint_dto.attributes, with_id),
        constraint_dto.resource_sets,
        with_id,
    )


def constraints_to_text(
    plain_dtos: Iterable[CibConstraintColocationDto],
    set_dtos: Iterable[CibConstraintColocationSetDto],
    with_id: bool,
) -> list[str]:
    result = []
    if plain_dtos:
        result.append("Colocation Constraints:")
        for constraint_dto in plain_dtos:
            result.extend(
                indent(
                    plain_constraint_to_text(constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    if set_dtos:
        result.append("Colocation Set Constraints:")
        for set_constraint_dto in set_dtos:
            result.extend(
                indent(
                    set_constraint_to_text(set_constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    return result


def _attributes_to_cmd_pairs(
    attributes_dto: CibConstraintColocationAttributesDto,
    filter_out: StringCollection = tuple(),
) -> Optional[list[tuple[str, str]]]:
    if attributes_dto.lifetime:
        warn(
            "Lifetime configuration detected in constraint "
            f"'{attributes_dto.constraint_id}' but not supported by this "
            "command."
            " Command for creating the constraint is omitted."
        )
        return None
    unsupported_options = {"influence"}
    result = []
    for pair in [("id", attributes_dto.constraint_id)] + _attributes_to_pairs(
        attributes_dto
    ):
        if pair[0] in unsupported_options:
            warn(
                f"Option '{pair[0]}' detected in constraint "
                f"'{attributes_dto.constraint_id}' but not supported by this "
                "command."
                " Command for creating the constraint is omitted."
            )
            return None
        if pair[0] in filter_out:
            continue
        result.append(pair)
    return result


def plain_constraint_to_cmd(
    constraint_dto: CibConstraintColocationDto,
) -> list[str]:
    if (
        constraint_dto.resource_instance is not None
        or constraint_dto.with_resource_instance is not None
    ):
        warn(
            "Resource instance(s) detected in constraint "
            f"'{constraint_dto.attributes.constraint_id}' but not supported by "
            "this command."
            " Command for creating the constraint is omitted."
        )
        return []
    if constraint_dto.node_attribute is not None:
        warn(
            "Option 'node-attribute' detected in constraint "
            f"'{constraint_dto.attributes.constraint_id}' but not supported by "
            "this command."
            " Command for creating the constraint is omitted."
        )
        return []
    result = [
        "pcs -- constraint colocation add {resource_role}{resource_id} with {with_resource_role}{with_resource_id}".format(
            resource_role=format_optional(constraint_dto.resource_role),
            resource_id=quote(constraint_dto.resource_id),
            with_resource_role=format_optional(
                constraint_dto.with_resource_role
            ),
            with_resource_id=quote(constraint_dto.with_resource_id),
        )
    ]
    pairs = _attributes_to_cmd_pairs(constraint_dto.attributes)
    if pairs is None:
        return []
    params = pairs_to_cmd(pairs)
    if params:
        result.extend(indent([params], indent_step=INDENT_STEP))
    return result


def set_constraint_to_cmd(
    constraint_dto: CibConstraintColocationSetDto,
) -> list[str]:
    result = ["pcs -- constraint colocation"]
    for resource_set in constraint_dto.resource_sets:
        set_cmd_part = _set.resource_set_to_cmd(resource_set)
        if not set_cmd_part:
            return []
        result.extend(indent(set_cmd_part, indent_step=INDENT_STEP))
    pairs = _attributes_to_cmd_pairs(constraint_dto.attributes)
    if pairs is None:
        return []
    params = pairs_to_cmd(pairs)
    if params:
        result.extend(indent([f"setoptions {params}"], indent_step=INDENT_STEP))
    return result



from textwrap import dedent
from unittest import TestCase

from pcs.cli.node import output as node_output
from pcs.common.pacemaker.node import CibNodeDto, CibNodeListDto
from pcs.common.pacemaker.nvset import CibNvpairDto, CibNvsetDto

from pcs_test.tools.custom_mock import RuleInEffectEvalMock
from pcs_test.tools.nodes_dto import get_nodes_dto

FIXTURE_NODE_CONFIG = get_nodes_dto(RuleInEffectEvalMock({}))


class NodeToLinesCmdBaseMixin:
    def assert_lines(self, config, output):
        self.assertEqual(
            "\n".join(self._call_func(config)) + "\n",
            output,
        )

    def test_nodes_without_nvpairs(self):
        config = CibNodeListDto(nodes=FIXTURE_NODE_CONFIG.nodes[2:2])
        output = "\n"
        self.assert_lines(config, output)


class NodeAttributeToLines(NodeToLinesCmdBaseMixin, TestCase):
    def _call_func(self, config):
        return node_output.config_dto_to_attribute_lines(config)

    def test_nodes_attributes_to_lines(self):
        config = FIXTURE_NODE_CONFIG
        output = dedent(
            """\
            Node: node1
              Description: node1 desc
              Attributes: nodes-1 score=50
                a=1
                b=2
                Rule: score=INFINITY
                  Expression: date lt 2000-01-01
            Node: node2
              Attributes: nodes-2
                a=1
                b=2
            """
        )
        self.assert_lines(config, output)


class NodeUtilizationToLines(NodeToLinesCmdBaseMixin, TestCase):
    def _call_func(self, config):
        return node_output.config_dto_to_utilization_lines(config)

    def test_nodes_utilization_to_lines(self):
        config = FIXTURE_NODE_CONFIG
        output = dedent(
            """\
            Node: node1
              Description: node1 desc
              Utilization: nodes-1-utilization score=50
                cpu=4
                ram=32
            Node: node2
              Utilization: nodes-2-utilization
                cpu=8
                ram=64
                Rule: boolean-op=and score=INFINITY
                  Expression: date gt 2000-01-01
            """
        )
        self.assert_lines(config, output)


class ConfigDtoToAttributeCmd(NodeToLinesCmdBaseMixin, TestCase):
    def _call_func(self, config):
        return node_output.config_dto_to_attribute_cmd(config)

    def test_config_dto_to_attribute_cmd(self):
        config = FIXTURE_NODE_CONFIG
        output = dedent(
            """\
            pcs -- node attribute node1 a=1 b=2
            pcs -- node attribute node2 a=1 b=2
            """
        )
        self.assert_lines(config, output)


class ConfigDtoToUtilizationCmd(NodeToLinesCmdBaseMixin, TestCase):
    def _call_func(self, config):
        return node_output.config_dto_to_utilization_cmd(config)

    def test_config_dto_to_utilization_cmd(self):
        config = FIXTURE_NODE_CONFIG
        output = dedent(
            """\
            pcs -- node utilization node1 cpu=4 ram=32
            pcs -- node utilization node2 cpu=8 ram=64
            """
        )
        self.assert_lines(config, output)


class FilterNodesByNodeName(TestCase):
    def test_no_match(self):
        self.assertEqual(
            CibNodeListDto(nodes=[]),
            node_output.filter_nodes_by_node_name(FIXTURE_NODE_CONFIG, "nodeX"),
        )

    def test_match(self):
        self.assertEqual(
            CibNodeListDto(nodes=FIXTURE_NODE_CONFIG.nodes[0:1]),
            node_output.filter_nodes_by_node_name(FIXTURE_NODE_CONFIG, "node1"),
        )


class FilterNodesByNvpairName(TestCase):
    nodes_config = CibNodeListDto(
        nodes=[
            CibNodeDto(
                id="1",
                uname="node1",
                description=None,
                score=None,
                type=None,
                instance_attributes=[
                    CibNvsetDto(
                        id="ia1",
                        options={},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(id="ia1b", name="b", value="2"),
                        ],
                    ),
                ],
                utilization=[
                    CibNvsetDto(
                        id="ua1",
                        options={},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(id="ua1a", name="a", value="1"),
                        ],
                    ),
                ],
            ),
            CibNodeDto(
                id="2",
                uname="node2",
                description=None,
                score=None,
                type=None,
                instance_attributes=[
                    CibNvsetDto(
                        id="ia2",
                        options={},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(id="ia2a", name="a", value="1"),
                        ],
                    ),
                ],
                utilization=[
                    CibNvsetDto(
                        id="ua2",
                        options={},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(id="ua2b", name="b", value="2"),
                        ],
                    ),
                ],
            ),
        ]
    )

    def test_no_nvpair_name_match(self):
        self.assertEqual(
            CibNodeListDto(
                nodes=[
                    CibNodeDto(
                        id="1",
                        uname="node1",
                        description=None,
                        score=None,
                        type=None,
                        instance_attributes=[
                            CibNvsetDto(
                                id="ia1",
                                options={},
                                rule=None,
                                nvpairs=[],
                            ),
                        ],
                        utilization=[
                            CibNvsetDto(
                                id="ua1",
                                options={},
                                rule=None,
                                nvpairs=[],
                            ),
                        ],
                    ),
                    CibNodeDto(
                        id="2",
                        uname="node2",
                        description=None,
                        score=None,
                        type=None,
                        instance_attributes=[
                            CibNvsetDto(
                                id="ia2",
                                options={},
                                rule=None,
                                nvpairs=[],
                            ),
                        ],
                        utilization=[
                            CibNvsetDto(
                                id="ua2",
                                options={},
                                rule=None,
                                nvpairs=[],
                            ),
                        ],
                    ),
                ]
            ),
            node_output.filter_nodes_nvpairs_by_name(self.nodes_config, "X"),
        )

    def test_match(self):
        self.assertEqual(
            CibNodeListDto(
                nodes=[
                    CibNodeDto(
                        id="1",
                        uname="node1",
                        description=None,
                        score=None,
                        type=None,
                        instance_attributes=[
                            CibNvsetDto(
                                id="ia1",
                                options={},
                                rule=None,
                                nvpairs=[],
                            ),
                        ],
                        utilization=[
                            CibNvsetDto(
                                id="ua1",
                                options={},
                                rule=None,
                                nvpairs=[
                                    CibNvpairDto(
                                        id="ua1a", name="a", value="1"
                                    ),
                                ],
                            ),
                        ],
                    ),
                    CibNodeDto(
                        id="2",
                        uname="node2",
                        description=None,
                        score=None,
                        type=None,
                        instance_attributes=[
                            CibNvsetDto(
                                id="ia2",
                                options={},
                                rule=None,
                                nvpairs=[
                                    CibNvpairDto(
                                        id="ia2a", name="a", value="1"
                                    ),
                                ],
                            ),
                        ],
                        utilization=[
                            CibNvsetDto(
                                id="ua2",
                                options={},
                                rule=None,
                                nvpairs=[],
                            ),
                        ],
                    ),
                ]
            ),
            node_output.filter_nodes_nvpairs_by_name(self.nodes_config, "a"),
        )



import json
from textwrap import dedent
from unittest import TestCase

from lxml import etree

from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
    ListCibNvsetDto,
)
from pcs.common.str_tools import format_list, format_plural

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import (
    PacemakerFeatures,
    get_tmp_file,
    write_data_to_tmpfile,
    write_file_to_tmpfile,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import XmlManipulation, etree_to_str

property_cib = rc("cib-property.xml")
UNCHANGED_CRM_CONFIG = etree_to_str(
    etree.parse(property_cib).findall(".//crm_config")[0]
)
FIXTURE_CONFIG_OUTPUT = dedent(
    """\
    Cluster Properties: cib-bootstrap-options
      cluster-name=HACluster
      enable-acl=false
      have-watchdog=false
      maintenance-mode=false
      placement-strategy=minimal
    """
)
_DEFAULT_MARK_REGEXP = r"(  .*=.* \(default\)\n)+"
_CONFIG_PROPERTIES_REGEXP_LIST = [
    r"Cluster Properties: cib-bootstrap-options\n",
    r"  cluster-name=HACluster\n",
    r"  enable-acl=false\n",
    r"  have-watchdog=false\n",
    r"  maintenance-mode=false\n",
    r"  placement-strategy=minimal\n",
]
CONFIG_ALL_REGEXP = (
    _DEFAULT_MARK_REGEXP.join(_CONFIG_PROPERTIES_REGEXP_LIST)
    + rf"{_DEFAULT_MARK_REGEXP}$"
)
DEFAULTS_REGEXP = r"^batch-limit=0\n(.*=.*\n)+$"
ALLOWED_PROPERTIES_REGEXP = r"'[^']+'(, '[^']+')*"


def get_invalid_option_messages_regexp(
    option_names, error=True, forceable=True
):
    error_occurred = (
        "Error: Errors have occurred, therefore pcs is unable to continue\n"
    )
    use_force = ", use --force to override"
    return (
        r"{severity}: invalid cluster property {option_pl} {option_name_list}, "
        r"allowed options are: {allowed_properties_regexp}{use_force}\n"
        r"{error_occurred}"
    ).format(
        severity="Error" if error else "Warning",
        option_name_list=format_list(option_names),
        option_pl=format_plural(option_names, "option", "options:"),
        allowed_properties_regexp=ALLOWED_PROPERTIES_REGEXP,
        use_force=use_force if error and forceable else "",
        error_occurred=error_occurred if error else "",
    )


class PropertyMixin(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(etree.parse(cib).findall(".//crm_config")[0])
    )
):
    def setUp(self):
        self.maxDiff = None
        self.temp_cib = get_tmp_file("tier1_cluster_property")
        write_file_to_tmpfile(property_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()


class TestPropertySet(PropertyMixin, TestCase):
    def test_success(self):
        self.assert_effect_single(
            (
                "property set enable-acl=true placement-strategy=utilization "
                "maintenance-mode="
            ).split(),
            """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-have-watchdog"
                        name="have-watchdog" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="HACluster"
                    />
                    <nvpair id="cib-bootstrap-options-placement-strategy"
                        name="placement-strategy" value="utilization"
                    />
                    <nvpair id="cib-bootstrap-options-enable-acl"
                        name="enable-acl" value="true"
                    />
                </cluster_property_set>
                <cluster_property_set id="second-set" score="10">
                    <nvpair id="second-set-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                </cluster_property_set>
            </crm_config>
            """,
        )

    def test_properties_to_set_missing(self):
        self.assert_pcs_fail(
            "property set".split(),
            stderr_start="\nUsage: pcs property set...",
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_missing_value(self):
        self.assert_pcs_fail(
            "property set keyword".split(),
            stderr_start="Error: missing value of 'keyword' option\n",
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unknown_properties(self):
        self.assert_pcs_fail(
            "property set unknown=value".split(),
            stderr_regexp=get_invalid_option_messages_regexp(["unknown"]),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unknown_properties_forced(self):
        self.assert_effect_single(
            "property set unknown=value --force".split(),
            """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-have-watchdog"
                        name="have-watchdog" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="HACluster"
                    />
                    <nvpair id="cib-bootstrap-options-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-placement-strategy"
                        name="placement-strategy" value="minimal"
                    />
                    <nvpair id="cib-bootstrap-options-enable-acl"
                        name="enable-acl" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-unknown" name="unknown"
                        value="value"
                    />
                </cluster_property_set>
                <cluster_property_set id="second-set" score="10">
                    <nvpair id="second-set-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                </cluster_property_set>
            </crm_config>
            """,
            stderr_regexp=get_invalid_option_messages_regexp(
                ["unknown"], error=False
            ),
        )

    def test_forbidden_properties(self):
        self.assert_pcs_fail(
            "property set cluster-name=NewName".split(),
            stderr_regexp=get_invalid_option_messages_regexp(
                ["cluster-name"], forceable=False
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_forbidden_properties_forced(self):
        self.assert_pcs_fail(
            "property set cluster-name=NewName --force".split(),
            stderr_regexp=get_invalid_option_messages_regexp(
                ["cluster-name"], forceable=False
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_set_stonith_watchdog_timeout_invalid_value(self):
        deprecation = ""
        if PacemakerFeatures.stonith_renamed_to_fencing():
            deprecation = (
                "Warning: cluster property option 'stonith-watchdog-timeout' "
                "is deprecated and might be removed in a future release, "
                "therefore it should not be used, use "
                "'fencing-watchdog-timeout' instead\n"
            )

        self.assert_pcs_fail(
            "property set stonith-watchdog-timeout=5x".split(),
            stderr_full=(
                deprecation
                + "Error: '5x' is not a valid stonith-watchdog-timeout value, "
                "use time interval (e.g. 1, 2s, 3m, 4h, ...)\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    @PacemakerFeatures.skip_unless_stonith_renamed_to_fencing()
    def test_set_fencing_watchdog_timeout_invalid_value(self):
        self.assert_pcs_fail(
            "property set fencing-watchdog-timeout=5x".split(),
            stderr_full=(
                "Error: '5x' is not a valid fencing-watchdog-timeout value, use"
                " time interval (e.g. 1, 2s, 3m, 4h, ...)\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_disable_fencing_warning(self):
        deprecation = ""
        if PacemakerFeatures.stonith_renamed_to_fencing():
            deprecation = (
                "Warning: cluster property option 'stonith-enabled' is "
                "deprecated and might be removed in a future release, therefore "
                "it should not be used, use 'fencing-enabled' instead\n"
            )
        self.assert_effect_single(
            "property set stonith-enabled=false".split(),
            """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-have-watchdog"
                        name="have-watchdog" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="HACluster"
                    />
                    <nvpair id="cib-bootstrap-options-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-placement-strategy"
                        name="placement-strategy" value="minimal"
                    />
                    <nvpair id="cib-bootstrap-options-enable-acl"
                        name="enable-acl" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-stonith-enabled"
                        name="stonith-enabled" value="false"
                    />
                </cluster_property_set>
                <cluster_property_set id="second-set" score="10">
                    <nvpair id="second-set-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                </cluster_property_set>
            </crm_config>

            """,
            stderr_full=deprecation
            + "Warning: Setting property stonith-enabled to false"
            " leaves the cluster with no enabled means to fence nodes,"
            " resulting in the cluster not being able to recover from"
            " certain failure conditions\n",
        )


class TestPropertyUnset(PropertyMixin, TestCase):
    def test_success(self):
        self.assert_effect_single(
            (
                "property unset placement-strategy enable-acl maintenance-mode"
            ).split(),
            """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-have-watchdog"
                        name="have-watchdog" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="HACluster"
                    />
                </cluster_property_set>
                <cluster_property_set id="second-set" score="10">
                    <nvpair id="second-set-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                </cluster_property_set>
            </crm_config>
            """,
        )

    def test_properties_to_set_missing(self):
        self.assert_pcs_fail(
            "property unset".split(),
            stderr_start="\nUsage: pcs property unset...",
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unset_not_configured_properties(self):
        self.assert_pcs_fail(
            "property unset missing1 missing2".split(),
            stderr_full=(
                "Error: Cannot remove properties 'missing1', 'missing2', they "
                "are not present in property set 'cib-bootstrap-options', use "
                "--force to override\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unset_not_configured_properties_forced(self):
        self.assert_effect_single(
            "property unset missing1 missing2 --force".split(),
            UNCHANGED_CRM_CONFIG,
            stderr_full=(
                "Warning: Cannot remove properties 'missing1', 'missing2', they "
                "are not present in property set 'cib-bootstrap-options'\n"
            ),
        )


class ConfigMixin(PropertyMixin):
    command = None

    def test_success(self):
        self.assert_pcs_success(self.command, stdout_full=FIXTURE_CONFIG_OUTPUT)

    def test_error(self):
        self.assert_pcs_fail(
            self.command + ["--output-format=format"],
            stderr_full=(
                "Error: Unknown value 'format' for '--output-format' option. "
                "Supported values are: 'cmd', 'json', 'text'\n"
            ),
        )

    def test_defaults_option(self):
        self.assert_pcs_success(
            self.command + ["--defaults"],
            stdout_regexp=DEFAULTS_REGEXP,
            stderr_full=(
                "Deprecation Warning: Option --defaults is deprecated and will "
                "be removed. Please use command 'pcs property defaults' "
                "instead.\n"
            ),
        )

    def test_all_option(self):
        self.assert_pcs_success(
            self.command + ["--all"], stdout_regexp=CONFIG_ALL_REGEXP
        )

    def test_json_format(self):
        stdout, stderr, retval = self.pcs_runner.run(
            "property config --output-format=json".split()
        )
        expected = ListCibNvsetDto(
            nvsets=[
                CibNvsetDto(
                    id="cib-bootstrap-options",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(
                            id="cib-bootstrap-options-cluster-name",
                            name="cluster-name",
                            value="HACluster",
                        ),
                        CibNvpairDto(
                            id="cib-bootstrap-options-enable-acl",
                            name="enable-acl",
                            value="false",
                        ),
                        CibNvpairDto(
                            id="cib-bootstrap-options-have-watchdog",
                            name="have-watchdog",
                            value="false",
                        ),
                        CibNvpairDto(
                            id="cib-bootstrap-options-maintenance-mode",
                            name="maintenance-mode",
                            value="false",
                        ),
                        CibNvpairDto(
                            id="cib-bootstrap-options-placement-strategy",
                            name="placement-strategy",
                            value="minimal",
                        ),
                    ],
                )
            ]
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_cmd_format_with_readonly_properties(self):
        self.assert_pcs_success(
            self.command + ["--output-format=cmd"],
            stdout_full=dedent(
                """\
                pcs property set --force -- \\
                  enable-acl=false \\
                  maintenance-mode=false \\
                  placement-strategy=minimal
                """
            ),
        )

    def _get_as_json(self, runner):
        stdout, stderr, retval = runner.run(
            self.command + ["--output-format=json"]
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def test_cmd_format_without_readonly_properties(self):
        orig_cib = get_tmp_file("tier1_cluster_property_orig")
        new_cib = get_tmp_file("tier1_cluster_property_new")
        xml_manip = XmlManipulation.from_file(rc("cib-empty.xml"))
        write_data_to_tmpfile(str(xml_manip), new_cib)
        xml_manip.append_to_first_tag_name(
            "crm_config",
            """
            <cluster_property_set id="cib-bootstrap-options">
                <nvpair id="cib-bootstrap-options-maintenance-mode"
                    name="maintenance-mode" value="false"/>
                <nvpair id="cib-bootstrap-options-placement-strategy"
                    name="placement-strategy" value="minimal"/>
                <nvpair id="cib-bootstrap-options-enable-acl" name="enable-acl"
                    value="false"/>
            </cluster_property_set>
            """,
        )
        write_data_to_tmpfile(str(xml_manip), orig_cib)
        pcs_runner_new = PcsRunner(new_cib.name)
        pcs_runner_orig = PcsRunner(orig_cib.name)

        stdout, stderr, retval = pcs_runner_orig.run(
            self.command + ["--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        self.assertEqual(stderr, "")

        cmd = stdout.replace("\\\n", "").split()
        stdout, stderr, retval = pcs_runner_new.run(cmd[1:])
        self.assertEqual(
            retval,
            0,
            (
                f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                f"stderr:\n{stderr}"
            ),
        )
        self.assertEqual(
            self._get_as_json(pcs_runner_orig),
            self._get_as_json(pcs_runner_new),
        )
        orig_cib.close()
        new_cib.close()


class TestProperty(ConfigMixin, TestCase):
    command = ["property"]


class TestPropertyConfig(ConfigMixin, TestCase):
    command = ["property", "config"]

    def test_specific_properties(self):
        self.assert_effect_single(
            self.command + ["maintenance-mode", "batch-limit", "nodefault"],
            UNCHANGED_CRM_CONFIG,
            stdout_full=dedent(
                """\
                Cluster Properties: cib-bootstrap-options
                  batch-limit=0 (default)
                  maintenance-mode=false
                """
            ),
        )


class TestPropertyDefaults(TestCase, AssertPcsMixin):
    def setUp(self):
        self.pcs_runner = PcsRunner(None)

    def test_success(self):
        self.assert_pcs_success(
            "property defaults".split(), stdout_regexp=DEFAULTS_REGEXP
        )

    def test_success_full(self):
        self.assert_pcs_success(
            "property defaults --full".split(), stdout_regexp=DEFAULTS_REGEXP
        )

    def test_success_specific_properties_also_advanced(self):
        self.assert_pcs_success(
            "property defaults no-quorum-policy stonith-enabled".split(),
            stdout_full=dedent(
                """\
                no-quorum-policy=stop
                stonith-enabled=true
                """
            ),
        )

    def test_notexistent(self):
        self.assert_pcs_fail(
            "property defaults nonexistent".split(),
            stderr_full="Error: No default value for property: 'nonexistent'\n",
        )

    def test_unsupported_option(self):
        self.assert_pcs_fail(
            "property defaults --force".split(),
            stderr_full=(
                "Error: Specified option '--force' is not supported in this "
                "command\n"
            ),
        )


FIXTURE_BATCH_LIMIT_DESC = (
    "batch-limit\n"
    "  Description: Maximum number of jobs that the cluster may "
    'execute in parallel across all nodes. The "correct" value '
    "will depend on the speed and load of your network and "
    "cluster nodes. If set to 0, the cluster will impose a "
    "dynamically calculated limit when any node has a high load.\n"
    "  Type: integer\n"
    "  Default: 0\n"
)

FIXTURE_CLUSTER_NAME_DESC = (
    "cluster-name\n"
    "  Description: An arbitrary name for the cluster. This optional value is "
    "mostly for users' convenience as desired in administration, but may also "
    "be used in Pacemaker configuration rules via the #cluster-name node "
    "attribute, and by higher-level tools and resource agents.\n"
    "  Type: string\n"
)


class TestPropertyDescribe(TestCase, AssertPcsMixin):
    def setUp(self):
        self.pcs_runner = PcsRunner(None)

    def test_success(self):
        self.assert_pcs_success(
            "property describe".split(),
            stdout_start=FIXTURE_BATCH_LIMIT_DESC,
        )

    def test_success_no_full(self):
        self.assert_pcs_success(
            "property describe".split(),
            stdout_regexp=r"(?!.*\(advanced use only\)).*",
        )

    def test_success_full(self):
        self.assert_pcs_success(
            "property describe --full".split(),
            stdout_regexp=r".*\(advanced use only\).*",
        )

    def test_success_specific(self):
        self.assert_pcs_success(
            "property describe cluster-name batch-limit".split(),
            stdout_full=FIXTURE_BATCH_LIMIT_DESC + FIXTURE_CLUSTER_NAME_DESC,
        )

    def test_notexistent(self):
        self.assert_pcs_fail(
            "property describe nonexistent".split(),
            stderr_full="Error: No description for property: 'nonexistent'\n",
        )

    def test_success_output_format_json(self):
        self.assert_pcs_success(
            "property describe --output-format=json".split(),
            stdout_regexp=(
                r'^{"properties_metadata": \[.*\], '
                r'"readonly_properties": \[.*\].*}$'
            ),
        )

    def test_fail_filter_and_output_format(self):
        self.assert_pcs_fail(
            "property describe property-name --output-format=json".split(),
            stderr_full=(
                "Error: property filtering is not supported with "
                "--output-format=json\n"
            ),
        )


class TestGetClusterPropertiesDefinition(AssertPcsMixin, TestCase):
    def setUp(self):
        self.pcs_runner = PcsRunner(None)

    def test_success(self):
        self.assert_pcs_success(
            "property get_cluster_properties_definition".split(),
            stdout_regexp='^{".*": {.*}.*}$',
        )

    def test_fail_bad_syntax(self):
        self.assert_pcs_fail(
            "property get_cluster_properties_definition arg".split(),
            stderr_start="\nUsage: pcs property ...",
        )



from typing import Mapping, Union

from pcs import settings
from pcs.common import reports
from pcs.common.pacemaker.cluster_property import ClusterPropertyMetadataDto
from pcs.common.pacemaker.nvset import ListCibNvsetDto
from pcs.common.str_tools import join_multilines
from pcs.common.types import StringSequence
from pcs.lib import cluster_property
from pcs.lib.cib import nvpair_multi, rule
from pcs.lib.cib.rule.in_effect import get_rule_evaluator
from pcs.lib.cib.tools import IdProvider, get_crm_config
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker.live import (
    get_cib_file_runner_env,
    has_cib_xml,
    is_crm_attribute_list_options_supported,
)
from pcs.lib.resource_agent import (
    ResourceAgentError,
    ResourceAgentFacade,
    ResourceAgentFacadeFactory,
    ResourceAgentMetadata,
    resource_agent_error_to_report_item,
)
from pcs.lib.resource_agent import const as ra_const


def _get_properties_facade(
    report_processor: reports.ReportProcessor,
    runner: CommandRunner,
) -> ResourceAgentFacade:
    if not is_crm_attribute_list_options_supported(runner):
        report_processor.report(
            reports.ReportItem.error(
                reports.messages.ClusterOptionsMetadataNotSupported()
            )
        )
        raise LibraryError()

    try:
        factory = ResourceAgentFacadeFactory(runner, report_processor)
        return factory.facade_from_crm_attribute(ra_const.CLUSTER_OPTIONS)
    except ResourceAgentError as e:
        report_processor.report_list(
            [
                resource_agent_error_to_report_item(
                    e, reports.ReportItemSeverity.error()
                )
            ]
        )
        raise LibraryError() from e


# backward compatibility layer - export cluster property metadata in the legacy
# format
def _cluster_property_metadata_to_dict(
    metadata: ResourceAgentMetadata,
) -> dict[str, dict[str, Union[bool, str, StringSequence]]]:
    banned_props = ["dc-version", "cluster-infrastructure"]
    basic_props = {
        "batch-limit": "Batch Limit",
        "no-quorum-policy": "No Quorum Policy",
        "symmetric-cluster": "Symmetric",
        "stonith-enabled": "Stonith Enabled",
        "fencing-enabled": "Fencing Enabled",
        "stonith-action": "Stonith Action",
        "cluster-delay": "Cluster Delay",
        "stop-orphan-resources": "Stop Orphan Resources",
        "stop-orphan-actions": "Stop Orphan Actions",
        "start-failure-is-fatal": "Start Failure is Fatal",
        "pe-error-series-max": "PE Error Storage",
        "pe-warn-series-max": "PE Warning Storage",
        "pe-input-series-max": "PE Input Storage",
        "enable-acl": "Enable ACLs",
    }
    property_definition = {}
    for parameter in metadata.parameters:
        if parameter.name in banned_props:
            continue
        single_property_dict: dict[str, Union[bool, str, StringSequence]] = {
            "name": parameter.name,
            "shortdesc": parameter.shortdesc or "",
            "longdesc": parameter.longdesc or "",
            "type": parameter.type,
            "default": parameter.default or "",
            "advanced": parameter.name not in basic_props or parameter.advanced,
            "readable_name": basic_props.get(parameter.name, parameter.name),
            "source": metadata.name.type,
        }
        if parameter.enum_values is not None:
            single_property_dict["enum"] = parameter.enum_values
            single_property_dict["type"] = "enum"
        property_definition[parameter.name] = single_property_dict
    return property_definition


def get_cluster_properties_definition_legacy(
    env: LibraryEnvironment,
) -> dict[str, dict[str, Union[bool, str, StringSequence]]]:
    """
    Return cluster properties definition in the legacy dictionary format.

    env -- provides communication with externals
    """
    return _cluster_property_metadata_to_dict(
        _get_properties_facade(env.report_processor, env.cmd_runner()).metadata
    )


def set_properties(
    env: LibraryEnvironment,
    cluster_properties: Mapping[str, str],
    force_flags: reports.types.ForceFlags = (),
) -> None:
    """
    Set specified pacemaker cluster properties, remove those with empty values.

    env -- provides communication with externals
    cluster_properties -- dictionary of cluster property names and values
    force_flags -- list of flags codes
    """
    cib = env.get_cib()
    runner = env.cmd_runner()
    id_provider = IdProvider(cib)
    force = reports.codes.FORCE in force_flags
    cluster_property_set_el = (
        cluster_property.get_cluster_property_set_element_legacy(
            cib, id_provider
        )
    )
    set_id = cluster_property_set_el.get("id", "")

    configured_properties = [
        nvpair_dto.name
        for nvpair_dto in nvpair_multi.nvset_element_to_dto(
            cluster_property_set_el,
            rule.RuleInEffectEvalDummy(),
        ).nvpairs
    ]

    env.report_processor.report_list(
        cluster_property.validate_set_cluster_properties(
            runner,
            _get_properties_facade(env.report_processor, runner),
            set_id,
            configured_properties,
            cluster_properties,
            env.service_manager,
            force=force,
        )
    )

    if env.report_processor.has_errors:
        raise LibraryError()

    nvpair_multi.nvset_update(
        cluster_property_set_el,
        id_provider,
        cluster_properties,
    )
    env.push_cib()


def get_properties(
    env: LibraryEnvironment, evaluate_expired: bool = False
) -> ListCibNvsetDto:
    """
    Get configured pacemaker cluster properties.

    env -- provides communication with externals
    evaluate_expired -- also evaluate whether rules are expired or in effect
    """
    cib = env.get_cib()
    rule_in_effect_eval = get_rule_evaluator(
        cib, env.cmd_runner(), env.report_processor, evaluate_expired
    )
    nvset_list = nvpair_multi.find_nvsets(
        get_crm_config(cib), nvpair_multi.NVSET_PROPERTY
    )
    return ListCibNvsetDto(
        nvsets=[
            nvpair_multi.nvset_element_to_dto(nvset_el, rule_in_effect_eval)
            for nvset_el in nvset_list
        ]
    )


def get_properties_metadata(
    env: LibraryEnvironment,
) -> ClusterPropertyMetadataDto:
    """
    Get pacemaker cluster properties metadata.

    Metadata is received in OCF 1.1 format from pacemaker daemons
    pacemaker-based, pacemaker-controld and pacemaker-schedulerd.

    env -- provides communication with externals
    """
    facade = _get_properties_facade(env.report_processor, env.cmd_runner())
    return ClusterPropertyMetadataDto(
        properties_metadata=[
            property_definition.to_dto()
            for property_definition in facade.metadata.parameters
        ],
        readonly_properties=cluster_property.READONLY_CLUSTER_PROPERTY_LIST,
    )


def remove_cluster_name(env: LibraryEnvironment) -> None:
    """
    Remove cluster-name property from CIB on local node. The cluster has to be
    stopped and the property is removed directly from the CIB file.
    """
    if env.service_manager.is_running("pacemaker"):
        env.report_processor.report(
            reports.ReportItem.error(reports.messages.PacemakerRunning())
        )
        raise LibraryError()

    if not has_cib_xml():
        env.report_processor.report(
            reports.ReportItem.error(reports.messages.CibXmlMissing())
        )
        raise LibraryError()

    xpath = "/cib/configuration/crm_config/cluster_property_set/nvpair[@name='cluster-name']"
    stdout, stderr, retval = env.cmd_runner().run(
        [settings.cibadmin_exec, "--delete-all", "--force", f"--xpath={xpath}"],
        env_extend=get_cib_file_runner_env(),
    )
    if retval != 0:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.CibClusterNameRemovalFailed(
                    reason=join_multilines([stderr, stdout])
                )
            )
        )
        raise LibraryError()



from pcs.cli.common.errors import CmdLineInputError


def qdevice_status_cmd(lib, argv, modifiers):
    """
    Options:
      * --full - get more detailed output
    """
    modifiers.ensure_only_supported("--full")
    if not argv or len(argv) > 2:
        raise CmdLineInputError()
    model = argv[0]
    cluster = None if len(argv) < 2 else argv[1]
    print(
        lib.qdevice.qdevice_status_text(
            model,
            verbose=modifiers.get("--full"),
            cluster=cluster,
        )
    )


def qdevice_setup_cmd(lib, argv, modifiers):
    """
    Options:
      * --enable - enable qdevice service
      * --start - start qdevice service
    """
    modifiers.ensure_only_supported("--enable", "--start")
    if len(argv) != 2:
        raise CmdLineInputError()
    if argv[0] != "model":
        raise CmdLineInputError()
    model = argv[1]
    lib.qdevice.qdevice_setup(
        model, modifiers.get("--enable"), modifiers.get("--start")
    )


def qdevice_destroy_cmd(lib, argv, modifiers):
    """
    Options:
      * --force - destroy qdevice even if it is used by clusters
    """
    modifiers.ensure_only_supported("--force")
    if len(argv) != 1:
        raise CmdLineInputError()
    model = argv[0]
    lib.qdevice.qdevice_destroy(model, proceed_if_used=modifiers.get("--force"))


def qdevice_start_cmd(lib, argv, modifiers):
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        raise CmdLineInputError()
    model = argv[0]
    lib.qdevice.qdevice_start(model)


def qdevice_stop_cmd(lib, argv, modifiers):
    """
    Options:
      * --force - stop qdevice even if it is used by clusters
    """
    modifiers.ensure_only_supported("--force")
    if len(argv) != 1:
        raise CmdLineInputError()
    model = argv[0]
    lib.qdevice.qdevice_stop(model, proceed_if_used=modifiers.get("--force"))


def qdevice_kill_cmd(lib, argv, modifiers):
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        raise CmdLineInputError()
    model = argv[0]
    lib.qdevice.qdevice_kill(model)


def qdevice_enable_cmd(lib, argv, modifiers):
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        raise CmdLineInputError()
    model = argv[0]
    lib.qdevice.qdevice_enable(model)


def qdevice_disable_cmd(lib, argv, modifiers):
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        raise CmdLineInputError()
    model = argv[0]
    lib.qdevice.qdevice_disable(model)



import json
from dataclasses import dataclass
from typing import Any, Iterable, Mapping, cast

from tornado.web import Finish

from pcs.common import file_type_codes, reports
from pcs.common.async_tasks import types
from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
)
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.common.reports.dto import ReportItemDto
from pcs.common.str_tools import format_list
from pcs.daemon.app.auth import LegacyTokenAuthenticationHandler
from pcs.daemon.async_tasks.scheduler import (
    Scheduler,
    TaskNotFoundError,
)
from pcs.daemon.async_tasks.types import Command
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.pcs_cfgsync.const import SYNCED_CONFIGS

from .common import RoutesType


@dataclass(frozen=True)
class SimplifiedResult:
    success: bool
    result: Any
    reports: list[ReportItemDto]


_SEVERITY_LABEL = {
    reports.ReportItemSeverity.DEBUG: "Debug: ",
    reports.ReportItemSeverity.DEPRECATION: "Deprecation warning: ",
    reports.ReportItemSeverity.ERROR: "Error: ",
    reports.ReportItemSeverity.INFO: "",
    reports.ReportItemSeverity.WARNING: "Warning: ",
}


def _report_to_str(report_item: ReportItemDto) -> str:
    return (
        _SEVERITY_LABEL.get(report_item.severity.level, "")
        + (f"{report_item.context.node}: " if report_item.context else "")
        + report_item.message.message
    )


def _reports_to_str(report_items: Iterable[ReportItemDto]) -> str:
    return "\n".join(_report_to_str(item) for item in report_items)


class _BaseApiV0Handler(LegacyTokenAuthenticationHandler):
    """
    Base class of handlers for the original API implemented in remote.rb
    """

    _scheduler: Scheduler

    def initialize(
        self, scheduler: Scheduler, auth_provider: AuthProvider
    ) -> None:
        # pylint: disable=arguments-differ
        super().initialize(auth_provider)
        self._scheduler = scheduler

    async def _handle_request(self) -> None:
        """
        Main method for handling requests
        """
        raise NotImplementedError()

    def _error(self, message: str, http_code: int = 400) -> Finish:
        """
        Helper method for exit request processing with an error
        """
        self.set_status(http_code)
        self.write(message)
        return Finish()

    def _check_required_params(self, required_params: set[str]) -> None:
        missing_params = required_params - set(self.request.arguments.keys())
        if missing_params:
            raise self._error(
                f"Required parameters missing: {format_list(missing_params)}"
            )

    async def _process_request(
        self, cmd_name: str, cmd_params: Mapping[str, Any]
    ) -> SimplifiedResult:
        """
        Helper method for calling pcs library commands
        """
        command_dto = CommandDto(
            command_name=cmd_name,
            params=cmd_params,
            options=CommandOptionsDto(
                effective_username=self.effective_user.username,
                effective_groups=list(self.effective_user.groups),
            ),
        )
        task_ident = self._scheduler.new_task(
            Command(command_dto, is_legacy_command=True),
            self.real_user,
        )

        try:
            task_result_dto = await self._scheduler.wait_for_task(
                task_ident, self.real_user
            )
        except TaskNotFoundError as e:
            raise self._error("Internal server error", 500) from e

        if (
            task_result_dto.task_finish_type == types.TaskFinishType.FAIL
            and task_result_dto.reports
            and task_result_dto.reports[0].message.code
            == reports.codes.NOT_AUTHORIZED
            and not task_result_dto.reports[0].context
        ):
            raise self._error("Permission denied", 403)

        if task_result_dto.task_finish_type == types.TaskFinishType.KILL:
            if (
                task_result_dto.kill_reason
                == types.TaskKillReason.COMPLETION_TIMEOUT
            ):
                raise self._error("Task processing timed out", 500)
            raise self._error("Task killed")

        if (
            task_result_dto.task_finish_type
            == types.TaskFinishType.UNHANDLED_EXCEPTION
        ):
            raise self._error("Unhandled exception", 500)

        return SimplifiedResult(
            task_result_dto.task_finish_type == types.TaskFinishType.SUCCESS,
            task_result_dto.result,
            task_result_dto.reports,
        )


class ResourceManageUnmanageHandler(_BaseApiV0Handler):
    async def _handle_request(self) -> None:
        self._check_required_params({"resource_list_json"})
        try:
            resource_list = json.loads(self.get_argument("resource_list_json"))
        except json.JSONDecodeError as e:
            raise self._error("Invalid input data format") from e
        result = await self._process_request(
            self._get_cmd(), dict(resource_or_tag_ids=resource_list)
        )
        if not result.success:
            raise self._error(_reports_to_str(result.reports))

    @staticmethod
    def _get_cmd() -> str:
        raise NotImplementedError()


class ResourceManageHandler(ResourceManageUnmanageHandler):
    @staticmethod
    def _get_cmd() -> str:
        return "resource.manage"


class ResourceUnmanageHandler(ResourceManageUnmanageHandler):
    @staticmethod
    def _get_cmd() -> str:
        return "resource.unmanage"


class QdeviceNetGetCaCertificateHandler(_BaseApiV0Handler):
    async def _handle_request(self) -> None:
        result = await self._process_request(
            "qdevice.qdevice_net_get_ca_certificate", {}
        )
        if not result.success:
            raise self._error(_reports_to_str(result.reports))
        self.write(result.result)


class QdeviceNetSignNodeCertificateHandler(_BaseApiV0Handler):
    async def _handle_request(self) -> None:
        self._check_required_params({"certificate_request", "cluster_name"})
        result = await self._process_request(
            "qdevice.qdevice_net_sign_certificate_request",
            dict(
                certificate_request=self.get_argument("certificate_request"),
                cluster_name=self.get_body_argument("cluster_name"),
            ),
        )
        if not result.success:
            raise self._error(_reports_to_str(result.reports))
        self.write(result.result)


class QdeviceNetClientInitCertificateStorageHandler(_BaseApiV0Handler):
    async def _handle_request(self) -> None:
        self._check_required_params({"ca_certificate"})
        result = await self._process_request(
            "qdevice.client_net_setup",
            dict(ca_certificate=self.get_argument("ca_certificate")),
        )
        if not result.success:
            raise self._error(_reports_to_str(result.reports))


class QdeviceNetClientImportCertificateHandler(_BaseApiV0Handler):
    async def _handle_request(self) -> None:
        self._check_required_params({"certificate"})
        result = await self._process_request(
            "qdevice.client_net_import_certificate",
            dict(certificate=self.get_argument("certificate")),
        )
        if not result.success:
            raise self._error(_reports_to_str(result.reports))


class QdeviceNetClientDestroyHandler(_BaseApiV0Handler):
    async def _handle_request(self) -> None:
        result = await self._process_request("qdevice.client_net_destroy", {})
        if not result.success:
            raise self._error(_reports_to_str(result.reports))


class GetConfigsHandler(_BaseApiV0Handler):
    _FILE_TYPE_CODE_TO_LEGACY = {
        file_type_codes.PCS_SETTINGS_CONF: "pcs_settings.conf",
        file_type_codes.PCS_KNOWN_HOSTS: "known-hosts",
    }

    async def _handle_request(self) -> None:
        result = await self._process_request(
            "pcs_cfgsync.get_configs",
            {"cluster_name": self.get_argument("cluster_name", "")},
        )

        if any(
            rep.message.code == reports.codes.FILE_IO_ERROR
            and rep.message.payload.get("file_type_code")
            == file_type_codes.COROSYNC_CONF
            and rep.message.payload.get("operation") == "read"
            for rep in result.reports
        ):
            # corosync.conf does not exists, or we were unable to read the file
            # so we say that the node is not in cluster - same as in the old
            # implementation
            self.write({"status": "not_in_cluster"})
            return
        if any(
            rep.message.code
            == reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME
            for rep in result.reports
        ):
            self.write({"status": "wrong_cluster_name"})
            return
        if not result.success or result.result is None:
            raise self._error(_reports_to_str(result.reports))

        command_result = cast(SyncConfigsDto, result.result)
        legacy_result = {
            "status": "ok",
            "cluster_name": command_result.cluster_name,
            "configs": {
                self._FILE_TYPE_CODE_TO_LEGACY[file_code]: {
                    "type": "file",
                    "text": command_result.configs.get(file_code),
                }
                for file_code in SYNCED_CONFIGS
            },
        }
        self.write(legacy_result)


def get_routes(scheduler: Scheduler, auth_provider: AuthProvider) -> RoutesType:
    def r(url: str) -> str:
        # pylint: disable=invalid-name
        return f"/remote/{url}"

    params = dict(scheduler=scheduler, auth_provider=auth_provider)
    return [
        # resources
        (r("manage_resource"), ResourceManageHandler, params),
        (r("unmanage_resource"), ResourceUnmanageHandler, params),
        # qdevice
        (
            r("qdevice_net_client_destroy"),
            QdeviceNetClientDestroyHandler,
            params,
        ),
        (
            # /api/v1/qdevice-client-net-import-certificate/v1
            r("qdevice_net_client_import_certificate"),
            QdeviceNetClientImportCertificateHandler,
            params,
        ),
        (
            r("qdevice_net_client_init_certificate_storage"),
            QdeviceNetClientInitCertificateStorageHandler,
            params,
        ),
        (
            r("qdevice_net_get_ca_certificate"),
            QdeviceNetGetCaCertificateHandler,
            params,
        ),
        (
            # /api/v1/qdevice-qdevice-net-sign-certificate-request/v1
            r("qdevice_net_sign_node_certificate"),
            QdeviceNetSignNodeCertificateHandler,
            params,
        ),
        # cfgsync
        (r("get_configs"), GetConfigsHandler, params),
    ]



import time
from typing import Any

from pcs import settings
from pcs.common.validate import is_integer
from pcs.lib.interface.config import FacadeInterface


class Facade(FacadeInterface):
    def __init__(self, parsed_config: dict[str, Any]):
        super().__init__(parsed_config)

    @classmethod
    def create(cls) -> "Facade":
        return cls({})

    @property
    def is_sync_allowed(self) -> bool:
        thread_disabled = self._config.get("thread_disabled", False)
        if not isinstance(thread_disabled, bool):
            thread_disabled = False
        return not (thread_disabled or self.is_sync_paused)

    @property
    def is_sync_paused(self) -> bool:
        paused_until = self._config.get("thread_paused_until")
        return (
            paused_until is not None
            and is_integer(paused_until)
            and time.time() < paused_until
        )

    @property
    def sync_interval(self) -> int:
        return self.__get_int(
            "thread_interval",
            settings.pcs_cfgsync_thread_interval_default,
            settings.pcs_cfgsync_thread_interval_minimum,
        )

    @property
    def sync_interval_previous_not_connected(self) -> int:
        return self.__get_int(
            "thread_interval_previous_not_connected",
            settings.pcs_cfgsync_thread_interval_previous_not_connected_default,
            settings.pcs_cfgsync_thread_interval_previous_not_connected_minimum,
        )

    @property
    def file_backup_count(self) -> int:
        return self.__get_int(
            "file_backup_count",
            settings.pcs_cfgsync_file_backup_count_default,
            settings.pcs_cfgsync_file_backup_count_minimum,
        )

    def __get_int(
        self, key: str, default_value: int, minimum_value: int
    ) -> int:
        real_value = self._config.get(key, default_value)
        if is_integer(real_value):
            return max(real_value, minimum_value)
        return default_value



import abc
from logging import Logger

from pcs.common.reports.utils import add_context_to_message

from .item import (
    ReportItem,
    ReportItemList,
    ReportItemSeverity,
)


class ReportProcessor(abc.ABC):
    def __init__(self) -> None:
        self._has_errors = False

    @property
    def has_errors(self) -> bool:
        return self._has_errors

    def report(self, report_item: ReportItem) -> "ReportProcessor":
        if _is_error(report_item):
            self._has_errors = True
        self._do_report(report_item)
        return self

    def report_list(self, report_list: ReportItemList) -> "ReportProcessor":
        for report_item in report_list:
            self.report(report_item)
        return self

    @abc.abstractmethod
    def _do_report(self, report_item: ReportItem) -> None:
        raise NotImplementedError()


def has_errors(report_list: ReportItemList) -> bool:
    return any(_is_error(report_item) for report_item in report_list)


def _is_error(report_item: ReportItem) -> bool:
    return report_item.severity.level == ReportItemSeverity.ERROR


class ReportProcessorToLog(ReportProcessor):
    def __init__(self, logger: Logger):
        super().__init__()
        self._logger = logger

    def _do_report(self, report_item: ReportItem) -> None:
        severity = report_item.severity.level

        context_dto = None
        if report_item.context:
            context_dto = report_item.context.to_dto()

        msg = add_context_to_message(report_item.message.message, context_dto)

        if severity == ReportItemSeverity.ERROR:
            self._logger.error(msg)
        elif severity in (
            ReportItemSeverity.WARNING,
            ReportItemSeverity.DEPRECATION,
        ):
            self._logger.warning(msg)
        elif severity == ReportItemSeverity.INFO:
            self._logger.info(msg)
        elif severity == ReportItemSeverity.DEBUG:
            self._logger.debug(msg)
        else:
            raise AssertionError("Unknown report severity")


class ReportProcessorInMemory(ReportProcessor):
    def __init__(self) -> None:
        super().__init__()
        self._reports: ReportItemList = []

    def _do_report(self, report_item: ReportItem) -> None:
        self._reports.append(report_item)

    @property
    def reports(self) -> ReportItemList:
        return self._reports
